import operator
from typing import AbstractSet

from pytest import fixture

from ...datalog.expression_processing import extract_logic_atoms
from ...datalog.translate_to_named_ra import TranslateToNamedRA
from ...datalog.wrapped_collections import WrappedRelationalAlgebraSet
from ...expressions import Constant, Symbol
from ...logic import Conjunction, Negation
from ...relational_algebra import Projection, Selection, int2columnint_constant
from ...relational_algebra.optimisers import (
    PushUnnamedSelectionsUp,
    RelationalAlgebraOptimiser
)
from ...relational_algebra.relational_algebra import RelationalAlgebraSolver
from ..shattering import sets_per_symbol, shatter_constants

RAO = RelationalAlgebraOptimiser()


EQ = Constant(operator.eq)
NE = Constant(operator.ne)

P = Symbol("P")
Q = Symbol("Q")
R = Symbol("R")
x = Symbol("x")
y = Symbol("y")
z = Symbol("z")
ans = Symbol("ans")
a = Constant("a")
b = Constant("b")
d = Constant("d")


def test_symbols_per_set_unitary():
    query = Conjunction((Q(x), Q(a), Q(b), P(x), Negation(Q(a))))

    res = sets_per_symbol(query)

    c0 = int2columnint_constant(0)
    expected = {
        Q: set([
            Projection(Selection(Q, EQ(c0, a)), (c0,)),
            Projection(Selection(Q, EQ(c0, b)), (c0,)),
            Projection(
                Selection(Q, Conjunction((NE(c0, a), NE(c0, b)))),
                (int2columnint_constant(0),)
            )
        ])
    }

    assert res == expected


def test_symbols_per_set_unitary_ne():
    query = Conjunction((Q(x), Q(a), Q(b), P(x), Negation(Q(a)), NE(d, x)))

    res = sets_per_symbol(query)

    c0 = int2columnint_constant(0)
    expected = {
        Q: set([
            Projection(Selection(Q, EQ(c0, a)), (c0,)),
            Projection(Selection(Q, EQ(c0, b)), (c0,)),
            Projection(Selection(Q, EQ(c0, d)), (c0,)),
            Projection(
                Selection(Q, Conjunction((NE(c0, a), NE(c0, b), NE(c0, d)))),
                (c0,)
            ),
        ]),
        P: set([
            Projection(Selection(P, EQ(c0, d)), (c0,)),
            Projection(
                Selection(P, NE(c0, d)),
                (c0,)
            )

        ])
    }

    assert res == expected


def test_symbols_per_set_binary():
    query = Conjunction((Q(x, b), Q(a, x), Q(b, y), NE(y, b)))

    res = sets_per_symbol(query)

    c0 = int2columnint_constant(0)
    c1 = int2columnint_constant(1)
    expected = {
        Q: set([
            Projection(
                Selection(Q, Conjunction((EQ(c0, a), EQ(c1, b)))), (c0, c1,)
            ),
            Projection(
                Selection(Q, Conjunction((EQ(c0, b), EQ(c1, b)))), (c0, c1,)
            ),
            Projection(
                Selection(Q, Conjunction((EQ(c0, a), NE(c1, b)))), (c0, c1,)
            ),
            Projection(
                Selection(Q, Conjunction((EQ(c0, b), NE(c1, b)))), (c0, c1,)
            ),
            Projection(
                Selection(Q, Conjunction((NE(c0, a), NE(c0, b), EQ(c1, b)))),
                (c0, c1)
            ),
            Projection(
                Selection(Q, Conjunction((NE(c0, a), NE(c0, b), NE(c1, b)))),
                (c0, c1)
            )
        ])
    }

    assert res == expected


@fixture
def R1():
    return Constant[AbstractSet](
        WrappedRelationalAlgebraSet([(chr(ord('a') + i),) for i in range(10)])
    )


@fixture
def R2():
    return Constant[AbstractSet](
        WrappedRelationalAlgebraSet([
            (chr(ord('a') + i * 2), i)
            for i in range(10)
        ])
    )


@fixture
def R3():
    return Constant[AbstractSet](
        WrappedRelationalAlgebraSet([
            (i,)
            for i in range(10)
        ])
    )


@fixture
def R4():
    return Constant[AbstractSet](
        WrappedRelationalAlgebraSet([
            (
                chr(ord('a') + i * 2),
                chr(ord('a') + i)
            )
            for i in range(10)
        ])
    )


class RelationalAlgebraPushInSolver(
    PushUnnamedSelectionsUp, RelationalAlgebraSolver
):
    pass


def test_shatter_unitary(R1, R2, R3):
    query = Conjunction((Q(x), Q(a), Q(b), R(y), Negation(P(b, y))))

    ra_query = TranslateToNamedRA().walk(query)

    symbol_table = dict({Q: R1, P: R2, R: R3})
    ras = RelationalAlgebraSolver(symbol_table)
    ra_sol = ras.walk(ra_query)

    res = shatter_constants(query, symbol_table, True)

    ras = RelationalAlgebraPushInSolver(symbol_table)
    ra_query = TranslateToNamedRA().walk(res)
    shattered_sol = ras.walk(ra_query)

    assert all(
        all(isinstance(arg, Symbol) for arg in atom.args)
        for atom in extract_logic_atoms(res)
    )
    assert ra_sol == shattered_sol


def test_shatter_binary_1(R4):
    query = Conjunction((Q(x, a), Q(y, a), Q(x, y)))
    ra_query = TranslateToNamedRA().walk(query)

    symbol_table = dict({Q: R4})
    ras = RelationalAlgebraSolver(symbol_table)
    ra_sol = ras.walk(ra_query)

    res = shatter_constants(query, symbol_table, True)

    ras = RelationalAlgebraPushInSolver(symbol_table)
    ra_query = TranslateToNamedRA().walk(res)
    shattered_sol = ras.walk(ra_query)

    assert all(
        all(isinstance(arg, Symbol) for arg in atom.args)
        for atom in extract_logic_atoms(res)
    )
    assert ra_sol == shattered_sol


def test_shatter_binary_2(R1, R4):
    query = Conjunction((Q(x, b), Q(a, x), Q(x, y), R(z)))
    ra_query = TranslateToNamedRA().walk(query)

    symbol_table = dict({Q: R4, R: R1})
    ras = RelationalAlgebraSolver(symbol_table)
    ra_sol = ras.walk(ra_query)

    res = shatter_constants(query, symbol_table, True)

    ras = RelationalAlgebraPushInSolver(symbol_table)
    ra_query = TranslateToNamedRA().walk(res)
    shattered_sol = ras.walk(ra_query)

    assert all(
        all(isinstance(arg, Symbol) for arg in atom.args)
        for atom in extract_logic_atoms(res)
    )
    assert ra_sol == shattered_sol


def test_shatter_binary_3(R4):
    query = Conjunction((Q(x, b), Q(y, y)))
    ra_query = TranslateToNamedRA().walk(query)

    symbol_table = dict({Q: R4})
    ras = RelationalAlgebraSolver(symbol_table)
    ra_sol = ras.walk(ra_query)

    res = shatter_constants(query, symbol_table, True)

    ras = RelationalAlgebraPushInSolver(symbol_table)
    ra_query = TranslateToNamedRA().walk(res)
    shattered_sol = ras.walk(ra_query)

    assert all(
        all(isinstance(arg, Symbol) for arg in atom.args)
        for atom in extract_logic_atoms(res)
    )
    assert ra_sol == shattered_sol


def test_shatter_binary_inequality(R1, R4):
    query = Conjunction((Q(x, a), Q(a, x), Q(y, y), NE(x, b), R(z)))
    ra_query = TranslateToNamedRA().walk(query)

    symbol_table = dict({Q: R4, R: R1})
    ras = RelationalAlgebraSolver(symbol_table)
    ra_sol = ras.walk(ra_query)

    res = shatter_constants(query, symbol_table, True)

    ras = RelationalAlgebraPushInSolver(symbol_table)
    ra_query = TranslateToNamedRA().walk(res)
    shattered_sol = ras.walk(ra_query)

    assert all(
        all(isinstance(arg, Symbol) for arg in atom.args)
        for atom in extract_logic_atoms(res)
    )
    assert ra_sol == shattered_sol



from neurolang.config import config
from .abstract import RelationalAlgebraColumnInt, RelationalAlgebraColumnStr

if "RAS" in config and (config["RAS"].get("backend") == "dask"):
    from .dask_sql import (
        NamedRelationalAlgebraFrozenSet,
        RelationalAlgebraFrozenSet,
        RelationalAlgebraSet,
        RelationalAlgebraStringExpression,
    )
else:
    from .pandas import (
        NamedRelationalAlgebraFrozenSet,
        RelationalAlgebraFrozenSet,
        RelationalAlgebraSet,
        RelationalAlgebraStringExpression,
    )


__all__ = [
    "RelationalAlgebraColumnInt",
    "RelationalAlgebraColumnStr",
    "RelationalAlgebraStringExpression",
    "RelationalAlgebraFrozenSet",
    "RelationalAlgebraSet",
    "NamedRelationalAlgebraFrozenSet",
]



r"""
Probabilistic Frontend
======================
Complements QueryBuilderDatalog class with probabilistic capabilities
1- add extensional probabilistic facts and choices
2- sove probabilistic queries
"""
import collections
import typing
from typing import (
    AbstractSet,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
)
from uuid import uuid1

import pandas as pd
from neurolang.type_system import (
    get_args,
    replace_type_variable_fix_python36_37,
)

from .. import expressions as ir
from ..datalog.aggregation import (
    BuiltinAggregationMixin,
    DatalogWithAggregationMixin,
    TranslateToLogicWithAggregation,
)
from ..datalog.chase import Chase
from ..datalog.constraints_representation import DatalogConstraintsProgram
from ..datalog.exceptions import InvalidMagicSetError
from ..datalog.expression_processing import (
    EqualitySymbolLeftHandSideNormaliseMixin,
)
from ..datalog.magic_sets import magic_rewrite
from ..datalog.negation import DatalogProgramNegationMixin
from ..datalog.ontologies_parser import OntologyParser
from ..datalog.ontologies_rewriter import OntologyRewriter

from ..exceptions import (
    UnsupportedQueryError,
    UnsupportedSolverError,
    UnsupportedProgramError,
)
from ..expression_walker import ExpressionBasicEvaluator, TypedSymbolTableMixin
from ..logic import Union, Symbol
from ..probabilistic import (
    dalvi_suciu_lift,
    small_dichotomy_theorem_based_solver,
)
from ..probabilistic.cplogic.program import CPLogicMixin
from ..probabilistic.expression_processing import (
    is_probabilistic_predicate_symbol,
    is_within_language_prob_query,
)
from ..probabilistic.magic_sets_processing import (
    probabilistic_postprocess_magic_rules,
)
from ..probabilistic.query_resolution import (
    QueryBasedProbFactToDetRule,
    compute_probabilistic_solution,
)
from ..probabilistic.stratification import (
    get_list_of_intensional_rules,
    stratify_program
)
from ..probabilistic.weighted_model_counting import (
    solve_marg_query as wmc_solve_marg_query,
)
from ..probabilistic.weighted_model_counting import (
    solve_succ_query as wmc_solve_succ_query,
)
from ..region_solver import RegionSolver
from ..relational_algebra import (
    NamedRelationalAlgebraFrozenSet,
    RelationalAlgebraColumnStr,
)
from ..commands import CommandsMixin
from ..datalog.basic_representation import UnionOfConjunctiveQueries
from . import query_resolution_expressions as fe
from .datalog.sugar import (
    TranslateProbabilisticQueryMixin,
    TranslateQueryBasedProbabilisticFactMixin,
)
from .datalog.sugar.spatial import TranslateEuclideanDistanceBoundMatrixMixin
from .datalog.syntax_preprocessing import ProbFol2DatalogMixin
from .frontend_extensions import NumpyFunctionsMixin
from .query_resolution_datalog import QueryBuilderDatalog


class RegionFrontendCPLogicSolver(
    EqualitySymbolLeftHandSideNormaliseMixin,
    TranslateProbabilisticQueryMixin,
    TranslateToLogicWithAggregation,
    TranslateQueryBasedProbabilisticFactMixin,
    TranslateEuclideanDistanceBoundMatrixMixin,
    QueryBasedProbFactToDetRule,
    ProbFol2DatalogMixin,
    RegionSolver,
    CommandsMixin,
    NumpyFunctionsMixin,
    CPLogicMixin,
    DatalogWithAggregationMixin,
    BuiltinAggregationMixin,
    DatalogProgramNegationMixin,
    DatalogConstraintsProgram,
    TypedSymbolTableMixin,
    ExpressionBasicEvaluator,
):
    pass


class NeurolangPDL(QueryBuilderDatalog):
    """
    Complements QueryBuilderDatalog class with probabilistic capabilities
    1- add extensional probabilistic facts and choices
    2- sove probabilistic queries
    """

    def __init__(
        self,
        chase_class: Type[Chase] = Chase,
        probabilistic_solvers: Tuple[Callable] = (
            small_dichotomy_theorem_based_solver.solve_succ_query,
            dalvi_suciu_lift.solve_succ_query,
            wmc_solve_succ_query,
        ),
        probabilistic_marg_solvers: Tuple[Callable] = (
            small_dichotomy_theorem_based_solver.solve_marg_query,
            dalvi_suciu_lift.solve_marg_query,
            wmc_solve_marg_query,
        ),
        check_qbased_pfact_tuple_unicity=False,
    ) -> "NeurolangPDL":
        """
        Query builder with probabilistic capabilities

        Parameters
        ----------
        chase_class : Type[Chase], optional
            used to compute deterministic solutions, by default Chase
        probabilistic_solvers : Tuple[Callable], optional
            used to compute probabilistic solutions,
            the order of the elements indicates the priority
            of usage of the solver.
            by default (lifted_solve_succ_query, wmc_solve_succ_query)

        probabilistic_marg_solvers : Tuple[Callable], optional
            used to compute probabilistic solutions,
            by default (lifted_solve_marg_query, wmc_solve_marg_query)


        Returns
        -------
        NeurolangPDL
            see description
        """
        super().__init__(
            RegionFrontendCPLogicSolver(), chase_class=chase_class
        )
        if len(probabilistic_solvers) != len(probabilistic_marg_solvers):
            raise ValueError(
                "probabilistic solve and marg "
                "solvers should be the same length"
            )
        self.probabilistic_solvers = probabilistic_solvers
        self.probabilistic_marg_solvers = probabilistic_marg_solvers
        self.current_program_rewritten = None
        self.check_qbased_pfact_tuple_unicity = (
            check_qbased_pfact_tuple_unicity
        )

    def load_ontology(
        self,
        paths: typing.Union[str, List[str]],
        connector_symbol_name=None
    ) -> None:
        """
        Loads and parses ontology stored at the specified paths, and
        store them into attributes

        Parameters
        ----------
        paths : typing.Union[str, List[str]]
            where the ontology files are stored
        connector_symbol_name : str
            name to be used in the connector_symbol.
            if None, the name is randomly generated

        Returns
        -------
        Expression
            the new connector_symbol, to be used in
            the query program.
        """
        if connector_symbol_name is None:
            self.connector_symbol = self.new_symbol()
        else:
            self.connector_symbol = self.new_symbol(name=connector_symbol_name)
        onto = OntologyParser(paths, connector_symbol=self.connector_symbol.expression)
        constraints, est_knowledge, entity_rules = onto.parse_ontology()
        self.program_ir.set_constraints(constraints)
        for name, expressions in constraints.items():
            symbol = Symbol(name=name).cast(UnionOfConjunctiveQueries)
            self.program_ir.symbol_table[symbol] = Union(expressions)
        for symbol, expressions in est_knowledge.items():
            iterable = [exp.args for exp in expressions]
            self.program_ir.add_extensional_predicate_from_tuples(
                symbol, iterable
            )
        self.program_ir.add_existential_rules(onto.existential_rules)
        for symbol, expressions in entity_rules.items():
            for e in expressions:
                self.program_ir.walk(e)

        return self.connector_symbol

    @property
    def current_program(self) -> List[fe.Expression]:
        """
        Returns the list of Front End Expressions that have
        currently been declared in the program, or through
        the program's constraints

        Returns
        -------
        List[fe.Expression]
            see description

        Example
        -------
        >>> nl = NeurolangPDL()
        >>> P = nl.add_uniform_probabilistic_choice_over_set(
        ...     [("a",), ("b",), ("c",)], name="P"
        ... )
        >>> Q = nl.add_uniform_probabilistic_choice_over_set(
        ...     [("a",), ("d",), ("c",)], name="Q"
        ... )
        >>> with nl.scope as e:
        ...     e.Z[e.x, e.PROB[e.x]] = P[e.x] & Q[e.x]
        ...     cp = nl.current_program
        >>> cp
        [
            Z(x, ( PROB(x) )) ← ( P(x) ) ∧ ( Q(x) )
        ]
        """
        cp = []
        for constraint in self.program_ir.constraints().formulas:
            cp.append(self.frontend_translator.walk(constraint))
        cp += super().current_program
        return cp

    def _execute_query(
        self,
        head: typing.Union[fe.Symbol, Tuple[fe.Expression, ...]],
        predicate: fe.Expression,
    ) -> Tuple[AbstractSet, Optional[ir.Symbol]]:
        """
        [Internal usage - documentation for developpers]

        Performs an inferential query: will return as first output
        an AbstractSet with as many elements as solutions
        of the predicate query. AbstractSet's columns correspond to
        the expressions in the head.
        Typically, probabilities are abstracted and processed similar
        to symbols, though of different nature (see examples)
        If head expressions are arguments of a functor, the latter will
        be returned as the second output, defaulted as None

        Parameters
        ----------
        head : typing.Union[
            fe.Symbol,
            Tuple[fe.Expression, ...],
        ]
            see description
        predicate : Expression
            see description

        Returns
        -------
        Tuple[AbstractSet, Optional[ir.Symbol]]
            see description

        Examples
        --------
        Note: example ran with pandas backend
        >>> nl = NeurolangPDL()
        >>> P = nl.add_uniform_probabilistic_choice_over_set(
        ...     [("a",), ("b",), ("c",)], name="P"
        ... )
        >>> Q = nl.add_uniform_probabilistic_choice_over_set(
        ...     [("a",), ("d",), ("c",)], name="Q"
        ... )
        >>> with nl.scope as e:
        ...     e.Z[e.x, e.PROB[e.x]] = P[e.x] & Q[e.x]
        ...     s1 = nl._execute_query(tuple(), e.Z[e.x, e.p])
        ...     s2 = nl._execute_query((e.x, ), e.Z[e.x, e.p])
        ...     s3 = nl._execute_query(e.Z[e.x, e.p], e.Z[e.x, e.p])
        >>> s1
        (
            C{
                Empty DataFrame
                Columns: []
                Index: [0, 1]
                : typing.AbstractSet
            },
            None
        )
        >>> s2
        (
            C{
                    x
                0   a
                1   c
                : typing.AbstractSet
            },
            None
        )
        >>> s3
        (
            C{
                    x   p
                0   a   0.111111
                1   c   0.111111
                : typing.AbstractSet
            },
            S{Z: Unknown}
        )
        """
        query_pred_symb = predicate.expression.functor
        query_pred_args = predicate.expression.args
        if is_probabilistic_predicate_symbol(query_pred_symb, self.program_ir):
            raise UnsupportedQueryError(
                "Queries on probabilistic predicates are not supported"
            )
        query = self.program_ir.symbol_table[query_pred_symb].formulas[0]

        try:
            with self.scope:
                goal, magic_rules = magic_rewrite(
                    query.consequent, self.program_ir
                )
                self.program_ir.walk(magic_rules)
                magic_query = self.program_ir.symbol_table[goal].formulas[0]
                (
                    magic_query,
                    magic_rules,
                ) = probabilistic_postprocess_magic_rules(
                    self.program_ir, magic_query, magic_rules
                )
            with self.scope:
                self.program_ir.walk(magic_rules)
                solution = self._solve(magic_query)
                query_pred_symb = magic_query.consequent.functor
        except (InvalidMagicSetError, UnsupportedProgramError) :
            solution = self._solve(query)

        if not isinstance(head, tuple):
            # assumes head is a predicate e.g. r(x, y)
            head_symbols = tuple(head.expression.args)
            functor_orig = head.expression.functor
        else:
            head_symbols = tuple(t.expression for t in head)
            functor_orig = None
        solution = self._restrict_to_query_solution(
            head_symbols, query_pred_symb, query_pred_args, solution
        )
        if functor_orig is None:
            solution = solution.value
        return solution, functor_orig

    def solve_all(self) -> Dict[str, NamedRelationalAlgebraFrozenSet]:
        """
        Returns a dictionary of "predicate_name": "Content"
        for all elements in the solution of the Datalog program.
        Typically, probabilities are abstracted and processed similar
        to symbols, though of different nature (see examples)

        Returns
        -------
        Dict[str, NamedRelationalAlgebraFrozenSet]
            extensional and intentional facts that have been derived
            through the current program, optionally with probabilities

        Example
        -------
        Note: example ran with pandas backend
        >>> nl = NeurolangPDL()
        >>> P = nl.add_uniform_probabilistic_choice_over_set(
        ...     [("a",), ("b",), ("c",)], name="P"
        ... )
        >>> Q = nl.add_uniform_probabilistic_choice_over_set(
        ...     [("a",), ("d",), ("c",)], name="Q"
        ... )
        >>> with nl.scope as e:
        ...     e.Z[e.PROB[e.x], e.x] = P[e.x] & Q[e.x]
        ...     solution = nl.solve_all()
        >>> solution
        {
            'Z':
                PROB        x
            0   0.111111    a
            1   0.111111    c
        }
        """
        solution_ir = self._solve()
        solution = {}
        for k, v in solution_ir.items():
            solution[k.name] = NamedRelationalAlgebraFrozenSet(
                self.predicate_parameter_names(k.name), v.value.unwrap()
            )
            solution[k.name].row_type = v.value.row_type
        return solution

    def _solve(self, query=None):
        idbs = stratify_program(query, self.program_ir)
        det_idb = idbs.get("deterministic", Union(tuple()))
        prob_idb = idbs.get("probabilistic", Union(tuple()))
        postprob_idb = idbs.get("post_probabilistic", Union(tuple()))
        solution = self._solve_deterministic_stratum(det_idb)
        if prob_idb.formulas:
            solution = self._solve_probabilistic_stratum(solution, prob_idb)
        if postprob_idb.formulas:
            solution = self._solve_postprobabilistic_deterministic_stratum(
                solution, postprob_idb
            )
        return solution

    def _solve_deterministic_stratum(self, det_idb):
        '''Resolution of the deterministic stratum. In case there
        are entries in the symbol table under the key __constraints__,
        a rewrite is performed and the resulting program is assigned
        to the variable `current_program_rewritten` to provide a way
        to access this information.

        In case connector_symbol has been defined, all rules derived
        from the ontology parsing are included in the program

        Parameters
        ----------
        det_idb : typing.Union
            union of rules composing the deterministic stratum.

        Returns
        -------
        Result obtained after resolution of the deterministic stratum
        '''
        if "__constraints__" in self.symbol_table:
            det_idb = self._rewrite_program_with_ontology(det_idb)
            if hasattr(self, 'connector_symbol'):
                connector_rules = tuple(
                    [q
                    for q in get_list_of_intensional_rules(self.program_ir)
                    if self.connector_symbol.expression in q.antecedent._symbols
                ])
                det_idb = Union(det_idb.formulas + connector_rules)
            self.current_program_rewritten = det_idb
        chase = self.chase_class(self.program_ir, rules=det_idb)
        solution = chase.build_chase_solution()
        return solution

    def _solve_probabilistic_stratum(self, solution, prob_idb):
        pfact_edb = self.program_ir.probabilistic_facts()
        pchoice_edb = self.program_ir.probabilistic_choices()
        for i, (succ_solver, marg_solver) in enumerate(
            zip(self.probabilistic_solvers, self.probabilistic_marg_solvers)
        ):
            try:
                prob_solution = compute_probabilistic_solution(
                    solution,
                    pfact_edb,
                    pchoice_edb,
                    prob_idb,
                    succ_solver,
                    marg_solver,
                    self.check_qbased_pfact_tuple_unicity,
                )
            except UnsupportedSolverError:
                if i == len(self.probabilistic_solvers) - 1:
                    raise
            else:
                break

        wlq_symbs = set(
            rule.consequent.functor
            for rule in prob_idb.formulas
            if is_within_language_prob_query(rule)
        )
        solution.update(
            {
                pred_symb: relation
                for pred_symb, relation in prob_solution.items()
                if pred_symb in wlq_symbs
            }
        )
        return solution

    def _solve_postprobabilistic_deterministic_stratum(
        self, solution, postprob_idb
    ):
        solver = RegionFrontendCPLogicSolver()
        for psymb, relation in solution.items():
            solver.add_extensional_predicate_from_tuples(
                psymb,
                relation.value,
            )
        for builtin_symb in self.program_ir.builtins():
            solver.symbol_table[builtin_symb] = self.program_ir.symbol_table[
                builtin_symb
            ]
        solver.walk(postprob_idb)
        chase = self.chase_class(solver, rules=postprob_idb)
        solution = chase.build_chase_solution()
        return solution

    @staticmethod
    def _restrict_to_query_solution(
        head_symbols, pred_symb, pred_args, solution
    ):
        """
        Based on a solution instance and a query predicate symbol and args,
        retrieve the relation whose columns correspond to symbols in the head
        of the query.
        """
        # return dum when empty solution (reported in GH481)
        if pred_symb not in solution:
            return ir.Constant[AbstractSet](
                NamedRelationalAlgebraFrozenSet.dum()
            )
        query_solution = solution[pred_symb].value.unwrap()
        query_row_type = solution[pred_symb].value.row_type
        constant_selection = {
            i: c.value for i, c in enumerate(pred_args)
            if isinstance(c, ir.Constant)
        }
        if constant_selection:
            query_solution = query_solution.selection(constant_selection)
        cols = list(
            arg.name if isinstance(arg, ir.Symbol)
            else ir.Symbol.fresh().name
            for arg in pred_args
        )
        query_solution = NamedRelationalAlgebraFrozenSet(cols, query_solution)
        query_solution = query_solution.projection(
            *(symb.name for symb in head_symbols)
        )
        type_args = get_args(query_row_type)
        proj_row_type = tuple(
            type_args[cols.index(symb.name)] for symb in head_symbols
        )
        query_solution.row_type = replace_type_variable_fix_python36_37(
            query_row_type, proj_row_type
        )
        return ir.Constant[AbstractSet](query_solution)

    def _rewrite_program_with_ontology(self, deterministic_program):
        orw = OntologyRewriter(
            deterministic_program,
            self.program_ir.get_constraints()
        )
        rewrite = orw.Xrewrite()

        return Union(rewrite)

    def add_probabilistic_facts_from_tuples(
        self,
        iterable: Iterable[Tuple[Any, ...]],
        type_: Type = ir.Unknown,
        name: Optional[str] = None,
    ) -> fe.Symbol:
        """
        Add probabilistic facts from tuples whose first element
        contains the probability label attached to that tuple.
        In the tuple (p, a, b, ...), p is the float probability
        of tuple (a, b, ...) to be True in any possible world.

        Note that each tuple from the iterable is independant
        from the others, meaning that multiple tuples can be True
        in the same possible world, contrary to a probabilistic choice.
        See example for details.

        Warning
        -------
        Typing for the iterable is improper, true -but yet unsupported
        in Python typing- typing should be Iterable[Tuple[float, Any, ...]]
        See examples

        Parameters
        ----------
        iterable : Iterable[Tuple[Any, ...]]
            the first float number represents the probability
            of the tuple constituted of the remaining elements
        type_ : Type, optional
            type for resulting AbstractSet if None will be inferred
            from the data, by default ir.Unknown
        name : Optional[str], optional
            name for the resulting fe.Symbol, if None
            will be fresh, by default None

        Returns
        -------
        fe.Symbol
            see description

        Example
        -------
        >>> nl = NeurolangPDL()
        >>> p = [(0.8, 'a', 'b'), (0.7, 'b', 'c')]
        >>> nl.add_probabilistic_facts_from_tuples(p, name="P")
        P: typing.AbstractSet[typing.Tuple[float, str]] = \
            [(0.8, 'a', 'b'), (0.7, 'b', 'c')]

        Adds the probabilistic facts:
            P((a, b)) : 0.8  <-  T
            P((b, c)) : 0.7  <-  T
        """
        return self._add_probabilistic_tuples(
            iterable,
            type_,
            name,
            self.program_ir.add_probabilistic_facts_from_tuples,
        )

    def add_probabilistic_choice_from_tuples(
        self,
        iterable: Iterable[Tuple[Any, ...]],
        type_: Type = ir.Unknown,
        name: Optional[str] = None,
    ) -> fe.Symbol:
        """
        Add probabilistic choice from tuples whose first element
        contains the probability label attached to that tuple.
        In the tuple (p, a, b, ...), p is the float probability
        of tuple (a, b, ...) to be True and all remaining tuples to
        be False in any possible world.

        Note that, contrary to a list of probabilistic facts, this represents
        a choice among possible values for the predicate, meaning that tuples
        in the set are mutually exclusive.
        See example for details.

        Warning
        -------
        Typing for the iterable is improper, true -but yet unsupported-
        typing should be Iterable[Tuple[float, Any, ...]]
        See examples

        Parameters
        ----------
        iterable : Iterable[Tuple[Any, ...]]
            the first float number represents the probability
            of the tuple constituted of the remaining elements
            Note that the float probabilities must sum to 1.
        type_ : Type, optional
            type for resulting AbstractSet if None will be inferred
            from the data, by default ir.Unknown
        name : Optional[str], optional
            name for the resulting fe.Symbol, if None
            will be fresh, by default None

        Returns
        -------
        fe.Symbol
            see description

        Raises
        ------
        DistributionDoesNotSumToOneError
            if float probabilities do not sum to 1.

        Example
        -------
        >>> nl = NeurolangPDL()
        >>> p = [(0.8, 'a', 'b'), (0.2, 'b', 'c')]
        >>> nl.add_probabilistic_choice_from_tuples(p, name="P")
        P: typing.AbstractSet[typing.Tuple[float, str, str]] = \
            [(0.8, 'a', 'b'), (0.2, 'b', 'c')]

        Adds the probabilistic choice:
            P((a, b)) : 0.8  v  P((b, c)) : 0.2  <-  T
        """
        return self._add_probabilistic_tuples(
            iterable,
            type_,
            name,
            self.program_ir.add_probabilistic_choice_from_tuples,
        )

    def _add_probabilistic_tuples(
        self, iterable, type_, name, solver_add_method
    ):
        if isinstance(iterable, pd.DataFrame):
            iterable = iterable.rename(
                columns={n: i for i, n in enumerate(iterable.columns)}
            )
        if name is None:
            name = str(uuid1())
        if isinstance(type_, tuple):
            type_ = Tuple[type_]
        symbol = ir.Symbol[AbstractSet[type_]](name)
        solver_add_method(symbol, iterable)
        return fe.Symbol(self, name)

    def add_uniform_probabilistic_choice_over_set(
        self,
        iterable: Iterable[Tuple[Any, ...]],
        type_: Type = ir.Unknown,
        name: Optional[str] = None,
    ) -> fe.Symbol:
        """
        Add uniform probabilistic choice over values in the iterable.

        Every tuple in the iterable will be assigned the same probability to
        be True, with all remaning tuples False, in any possible world.

        Note that, contrary to a list of probabilistic facts, this represents
        a choice among possible values for the predicate, meaning that tuples
        in the set are mutually exclusive.

        See example for details.

        Parameters
        ----------
        iterable : Iterable[Tuple[Any, ...]]
            typically List[Tuple[Any, ...]], other types of Iterable
            will be cast as lists
        type_ : Type, optional
            type for resulting AbstractSet if None will be inferred
            from the data, by default ir.Unknown
        name : Optional[str], optional
            name for the resulting fe.Symbol, if None
            will be fresh, by default None

        Returns
        -------
        fe.Symbol
            see description

        Example
        -------
        >>> nl = NeurolangPDL()
        >>> p = [('a',), ('b',)]
        >>> nl.add_uniform_probabilistic_choice_over_set(p, name="P")
        P: typing.AbstractSet[typing.Tuple[float, str]] = \
            [(0.5, 'a'), (0.5, 'b')]

        Adds the probabilistic choice:
            P(a) : 0.5  v  P(b) : 0.5  <-  T
        """
        if name is None:
            name = str(uuid1())
        if isinstance(type_, tuple):
            type_ = Tuple[type_]
        if isinstance(iterable, pd.DataFrame):
            iterable = iterable.rename(
                columns={n: i for i, n in enumerate(iterable.columns)}
            )
            arity = len(iterable.columns)
        else:
            arity = len(next(iter(iterable)))
        symbol = ir.Symbol[AbstractSet[type_]](name)
        columns = tuple(ir.Symbol.fresh().name for _ in range(arity))
        ra_set = NamedRelationalAlgebraFrozenSet(columns, iterable)
        prob_col = ir.Symbol.fresh().name
        probability = 1 / len(ra_set)
        projections = collections.OrderedDict()
        projections[prob_col] = probability
        for col in columns:
            projections[col] = RelationalAlgebraColumnStr(col)
        ra_set = ra_set.extended_projection(projections)
        self.program_ir.add_probabilistic_choice_from_tuples(
            symbol,
            ra_set.projection_to_unnamed(*projections.keys())
        )
        return fe.Symbol(self, name)



from .general import (
    ChaseGeneral,
    ChaseNaive,
    ChaseNode,
    ChaseNonRecursive,
    ChaseSemiNaive,
    ChaseStratified,
    NeuroLangNonLinearProgramException,
    NeuroLangProgramHasLoopsException,
)
from .mgu import ChaseMGUMixin
from .relational_algebra import (
    ChaseNamedRelationalAlgebraMixin,
    ChaseRelationalAlgebraPlusCeriMixin,
)
from ..aggregation import ChaseAggregationMixin

__all__ = [
    "ChaseGeneral",
    "ChaseNode",
    "ChaseNaive",
    "ChaseSemiNaive",
    "ChaseStratified",
    "NeuroLangNonLinearProgramException",
    "ChaseMGUMixin",
    "ChaseNonRecursive",
    "ChaseRelationalAlgebraPlusCeriMixin",
    "ChaseNamedRelationalAlgebraMixin",
    "NeuroLangProgramHasLoopsException",
]


class ChaseSN(ChaseSemiNaive, ChaseNamedRelationalAlgebraMixin, ChaseGeneral):
    pass


class ChaseNR(
    ChaseNonRecursive, ChaseNamedRelationalAlgebraMixin, ChaseGeneral
):
    pass


class ChaseN(ChaseNaive, ChaseNamedRelationalAlgebraMixin, ChaseGeneral):
    pass


class ChaseAggregationNR(ChaseAggregationMixin, ChaseNR):
    pass


class ChaseAggregationSN(ChaseAggregationMixin, ChaseSN):
    pass


class ChaseAggregationN(ChaseAggregationMixin, ChaseN):
    pass


DEFAULT_STRATIFIED_CHASE_CLASSES = (ChaseNR, ChaseSN, ChaseN)


class ChaseNamedStratified(ChaseStratified, ChaseNamedRelationalAlgebraMixin):
    def __init__(
        self,
        datalog_program,
        rules=None,
        chase_classes=DEFAULT_STRATIFIED_CHASE_CLASSES,
    ):
        super().__init__(datalog_program, chase_classes, rules=rules)


class Chase(ChaseNamedStratified):
    """
    Stratified Chase which will try to run (Aggregation + ChaseNonRecursive),
    (Aggregation + ChaseSemiNaive) and then (Aggregation + ChaseNaive)
    for each stratum.
    """

    def __init__(self, datalog_program, rules=None):
        chase_classes = (
            ChaseAggregationNR,
            ChaseAggregationSN,
            ChaseAggregationN,
        )
        super().__init__(
            datalog_program, rules=rules, chase_classes=chase_classes
        )



from operator import add, eq, ge, gt, le, lt, mul, ne, pow, sub, truediv

from lark import Lark
from lark import Transformer
from lark.exceptions import UnexpectedToken, UnexpectedCharacters, LarkError

from neurolang.logic import ExistentialPredicate

from ...datalog import Conjunction, Fact, Implication, Negation, Union
from ...datalog.constraints_representation import RightImplication
from ...expressions import (
    Command,
    Constant,
    Expression,
    FunctionApplication,
    Lambda,
    Query,
    Statement,
    Symbol
)
from ...probabilistic.expressions import (
    PROB,
    Condition,
    ProbabilisticFact
)
from ...exceptions import UnexpectedTokenError, UnexpectedCharactersError, NeuroLangException

GRAMMAR = u"""
start: expressions
expressions : (expression)+

?expression : rule
            | constraint
            | fact
            | probabilistic_rule
            | probabilistic_fact
            | statement
            | statement_function
            | command

probabilistic_rule : head "::" arithmetic_operation IMPLICATION (condition | body)

rule : (head | query) IMPLICATION (condition | body)
IMPLICATION : ":-" | "\N{LEFTWARDS ARROW}"

query : "ans(" [ arguments ] ")"

condition : composite_predicate "//" composite_predicate
?composite_predicate : "(" conjunction ")"
                        | predicate

constraint : body RIGHT_IMPLICATION head
RIGHT_IMPLICATION : "-:" | "\N{RIGHTWARDS ARROW}"

existential_predicate : exists "(" existential_body ")"
?existential_body : arguments SUCH_THAT predicate ( "," predicate )*
                  | arguments SUCH_THAT predicate ( "&" predicate )*

SUCH_THAT : "st" | ";"

?head : head_predicate
head_predicate : identifier "(" [ arguments ] ")"

?body : conjunction
conjunction : predicate ("," predicate)*
            | predicate ("&" predicate)*

negated_predicate : ("~" | "\u00AC" ) predicate
predicate : id_application
          | negated_predicate
          | existential_predicate
          | "(" comparison ")"
          | logical_constant

id_application : int_ext_identifier "(" [ arguments ] ")"

comparison : argument COMPARISON_OPERATOR argument
COMPARISON_OPERATOR : "==" | "<" | "<=" | ">=" | ">" | "!="

statement : identifier ":=" arithmetic_operation
statement_function : identifier "(" [ arguments ] ")" ":=" argument

probabilistic_fact : ( arithmetic_operation | int_ext_identifier ) "::" constant_predicate

command : "." cmd_identifier "(" [ cmd_args ] ")"
cmd_args : cmd_arg ("," cmd_arg)*
?cmd_arg : pos_item
         | keyword_item

keyword_item : identifier "=" pos_item
pos_item : arithmetic_operation | python_string

python_string : PYTHON_STRING
PYTHON_STRING : DOUBLE_QUOTE NO_DBL_QUOTE_STR DOUBLE_QUOTE
              | SINGLE_QUOTE NO_DBL_QUOTE_STR SINGLE_QUOTE
NO_DBL_QUOTE_STR : /[^"]*/

?function_application : "(" lambda_expression ")" "(" [ arguments ] ")" -> lambda_application
                     | id_application

signed_int_ext_identifier : int_ext_identifier     -> signed_int_ext_identifier
                          | "-" int_ext_identifier -> minus_signed_id
?int_ext_identifier : identifier
                    | ext_identifier
                    | lambda_expression

lambda_expression : "lambda" arguments ":" argument

arguments : argument ("," argument)*
argument : arithmetic_operation | DOTS
DOTS : "..."

arithmetic_operation : term                          -> sing_op
                     | arithmetic_operation "+" term -> plus_op
                     | arithmetic_operation "-" term -> minus_op
term : factor          -> sing_term
     | term "*" factor -> mul_term
     | term "/" factor -> div_term
factor : exponent             -> sing_factor
       | factor "**" exponent -> pow_factor
?exponent : literal | function_application | signed_int_ext_identifier | "(" argument ")"

fact : constant_predicate
constant_predicate : identifier "(" (literal | ext_identifier) ("," (literal | ext_identifier))* ")"
                   | identifier "(" ")"

?literal : number | text

ext_identifier : "@" identifier
identifier : cmd_identifier | identifier_regexp
identifier_regexp : IDENTIFIER_REGEXP
IDENTIFIER_REGEXP : "`" /[0-9a-zA-Z\/#%\._:-]+/ "`"
cmd_identifier : CMD_IDENTIFIER
CMD_IDENTIFIER : /\\b(?!\\bexists\\b)(?!\\b\\u2203\\b)(?!\\bEXISTS\\b)(?!\\bst\\b)(?!\\bans\\b)[a-zA-Z_][a-zA-Z0-9_]*\\b/

exists : EXISTS
EXISTS : "exists" | "\u2203" | "EXISTS"

text : TEXT
TEXT : DOUBLE_QUOTE /[a-zA-Z0-9 ]*/ DOUBLE_QUOTE
     | SINGLE_QUOTE /[a-zA-Z0-9 ]*/ SINGLE_QUOTE
DOUBLE_QUOTE : "\\""
SINGLE_QUOTE : "'"

?number : integer | float
?integer : INT     -> pos_int
         | "-" INT -> neg_int
?float : FLOAT     -> pos_float
       | "-" FLOAT -> neg_float

logical_constant : FALSE | TRUE
TRUE             : "True" | "\u22A4"
FALSE            : "False" | "\u22A5"

WHITESPACE : /[\t ]+/
//NEWLINE : /[(\\u000C)? (\\r)? \\n]+/

%import common.INT
%import common.FLOAT
%import common.NEWLINE

%ignore WHITESPACE
%ignore NEWLINE
"""

OPERATOR = {
    "+": add,
    "-": sub,
    "==": eq,
    ">=": ge,
    ">": gt,
    "<=": le,
    "<": lt,
    "*": mul,
    "!=": ne,
    "**": pow,
    "/": truediv,
}


COMPILED_GRAMMAR = Lark(GRAMMAR, parser='lalr', debug=True)


class ExternalSymbol(Symbol):
    def __repr__(self):
        return "@S{{{}: {}}}".format(self.name, self.__type_repr__)


class DatalogTransformer(Transformer):

    def __init__(self, locals=None, globals=None):
        super().__init__()

        if locals is None:
            locals = {}
        if globals is None:
            globals = {}

        self.locals = locals
        self.globals = globals

    def start(self, ast):
        return ast[0]

    def expressions(self, ast):
        if isinstance(ast[0], Expression):
            return Union(ast)
        else:
            ast = ast[0]
        return Union(ast)

    #     return 0

    def probabilistic_rule(self, ast):
        head = ast[0]
        probability = ast[1]
        body = ast[3]
        return Implication(
            ProbabilisticFact(probability, head),
            body,
        )

    def rule(self, ast):
        head = ast[0]
        if isinstance(head, Expression) and head.functor == Symbol("ans"):
            return Query(ast[0], ast[2])
        else:
            return Implication(ast[0], ast[2])

    def condition(self, ast):
        conditioned = ast[0]
        condition = ast[1]
        return Condition(conditioned, condition)

    def constraint(self, ast):
        return RightImplication(ast[0], ast[2])

    def conjunction(self, ast):
        conj = []
        for c in ast:
            if isinstance(c, Expression):
                conj.append(c)
        return Conjunction(conj)

    def existential_predicate(self, ast):
        ast1 = ast[1].children
        exp = []
        for i in ast1:
            if isinstance(i, FunctionApplication):
                exp.append(i)

        if len(exp) == 1:
            exp = exp[0]
        else:
            exp = Conjunction(tuple(exp))

        for arg in ast1[0]:
            exp = ExistentialPredicate(arg, exp)

        return exp

    def negated_predicate(self, ast):
        return Negation(ast[0])

    def comparison(self, ast):
        operator = Constant(OPERATOR[ast[1].value])
        return operator(ast[0], ast[2])

    def predicate(self, ast):
        if not isinstance(ast, Expression):
            if isinstance(ast, list):
                if len(ast) == 1:
                    ast = ast[0]
                else:
                    ast = ast[0](*ast[1])
        return ast

    def head_predicate(self, ast):
        if ast[1] != None:
            arguments = list(ast[1])

            if PROB in arguments:
                ix_prob = arguments.index(PROB)
                arguments_not_prob = (
                        arguments[:ix_prob] +
                        arguments[ix_prob + 1:]
                )
                prob_arg = PROB(*arguments_not_prob)
                arguments = (
                        arguments[:ix_prob] +
                        [prob_arg] +
                        arguments[ix_prob + 1:]
                )

            ast = ast[0](*arguments)
        else:
            ast = ast[0]()
        return ast

    def query(self, ast):
        ast = ast[0]
        if ast != None:
            if isinstance(ast, tuple):
                arguments = ast
            else:
                arguments = tuple(ast)

            return Symbol("ans")(*arguments)
        else:
            return Symbol("ans")()

    def statement(self, ast):
        return Statement(ast[0], ast[1])

    def statement_function(self, ast):
        return Statement(
            ast[0],
            Lambda(ast[1], ast[2])
        )

    def probabilistic_fact(self, ast):
        return Implication(
            ProbabilisticFact(ast[0], ast[1]),
            Constant(True),
        )

    def command(self, ast):
        if ast[1] == None:
            cmd = Command(ast[0], (), ())
        else:
            # only args, only kwargs or both
            name = ast[0]
            args, kwargs = ast[1]
            cmd = Command(name, args, kwargs)
        return cmd

    def cmd_args(self, ast):
        args = ()
        kwargs = ()
        for a in ast :
            if isinstance(a, tuple):
                kwargs = kwargs + (a,)
            else:
                args = args + (a,)
        return args, kwargs

    def keyword_args(self, ast):
        return tuple(ast)

    def keyword_item(self, ast):
        key = ast[0]
        return (key, ast[1])

    def pos_args(self, ast):
        return tuple(ast)

    def pos_item(self, ast):
        ast = ast[0]
        if not isinstance(ast, Expression):
            return Constant((ast.children[0]).value.replace('"', ''))
        return ast

    def lambda_application(self, ast):
        if not isinstance(ast[0], Expression):
            f = Symbol(ast[0])
        else:
            f = ast[0]

        return FunctionApplication(f, args=ast[1])

    def id_application(self, ast):
        if not isinstance(ast[0], Expression):
            f = Symbol(ast[0])
        else:
            f = ast[0]

        return FunctionApplication(f, args=ast[1])

    def minus_signed_id(self, ast):
        return Constant(mul)(Constant(-1), ast[0])

    def signed_int_ext_identifier(self, ast):
        return ast[0]

    def lambda_expression(self, ast):
        return Lambda(ast[0], ast[1])

    def arguments(self, ast):
        return tuple(ast)

    def argument(self, ast):
        ast = ast[0]
        if isinstance(ast, Expression):
            return ast
        else:
            return Symbol.fresh()

    def minus_op(self, ast):
        if len(ast) == 1:
            if isinstance(ast[0], Expression):
                return ast[0]

            else:
                return ast[0]

        op_str = "-"
        op = Constant(OPERATOR[op_str])
        return op(*ast)

    def plus_op(self, ast):
        if len(ast) == 1:
            if isinstance(ast[0], Expression):
                return ast[0]
            else:
                return ast[0]

        op_str = "+"
        op = Constant(OPERATOR[op_str])
        return op(*ast)

    def sing_op(self, ast):
        if len(ast) == 1:
            if isinstance(ast[0], Expression):
                return ast[0]

            else:
                return ast[0]
        return 0

    def term(self, ast):
        if isinstance(ast, Expression):
            return ast
        elif len(ast) == 1:
            return ast[0]

    def div_term(self, ast):
        if len(ast) == 1:
            if isinstance(ast[0], Expression):
                return ast[0]
            else:
                return 0

        op_str = "/"
        op = Constant(OPERATOR[op_str])

        return op(ast[0], ast[1])

    def mul_term(self, ast):
        if len(ast) == 1:
            if isinstance(ast[0], Expression):
                return ast[0]
            else:
                return 0

        op_str = "*"
        op = Constant(OPERATOR[op_str])

        return op(ast[0], ast[1])

    def sing_term(self, ast):
        if len(ast) == 1:
            if isinstance(ast[0], Expression):
                return ast[0]
            else:
                return 0
        else:
            return 0

    def factor(self, ast):
        if isinstance(ast[0], Expression):
            return ast[0]
        elif len(ast) == 1:
            return 0
        else:
            return 0

    def pow_factor(self, ast):
        if len(ast) == 1:
            if isinstance(ast[0], Expression):
                return ast[0]
            else:
                return 0
        else:
            return Constant(pow)(ast[0], ast[1])

    def sing_factor(self, ast):
        if len(ast) == 1:
            if isinstance(ast[0], Expression):
                return ast[0]
            else:
                return 0
        else:
            return 0

    def fact(self, ast):
        return Fact(ast[0])

    def constant_predicate(self, ast):
        predicate_name = ast[0]
        del ast[0]
        return predicate_name(*ast)

    def ext_identifier(self, ast):
        ast = ast[0]
        return ExternalSymbol[ast.type](ast.name)

    def identifier(self, ast):
        ast = ast[0]
        if not isinstance(ast, Symbol):
            return Symbol(ast)

        return ast

    def identifier_regexp(self, ast):
        return (ast[0]).value.replace('`', '')

    def cmd_identifier(self, ast):
        return Symbol((ast[0]).value)

    def text(self, ast):
        return Constant((ast[0].replace("'", "")).replace('"', ''))

    def pos_int(self, ast):
        return Constant(int((ast[0]).value))

    def neg_int(self, ast):
        return Constant(0 - int((ast[0]).value))

    def pos_float(self, ast):
        return Constant(float((ast[0]).value))

    def neg_float(self, ast):
        return Constant(0 - float((ast[0]).value))

    def _default(self, ast):
        return ast


def parser(code, locals=None, globals=None):

    try:
        jp = COMPILED_GRAMMAR.parse(code.strip())
        return DatalogTransformer().transform(jp)
    except UnexpectedToken as e:
        raise UnexpectedTokenError(str(e), line=e.line - 1, column=e.column - 1) from e
    except UnexpectedCharacters as e:
        raise UnexpectedCharactersError(str(e), line=e.line - 1, column=e.column - 1) from e
    except LarkError as e:
        raise NeuroLangException from e



import inspect
import operator as op
from typing import AbstractSet, Callable, Mapping, Sequence, Tuple

import pickle
import pytest
import numpy


from ..config import config
from .. import expressions, logic
from ..expression_walker import (
    ExpressionBasicEvaluator,
    TypedSymbolTableEvaluator,
)
from ..expressions import Expression, Unknown, expressions_behave_as_objects

C_ = expressions.Constant
S_ = expressions.Symbol
F_ = expressions.FunctionApplication
L_ = expressions.Lambda
E_ = logic.ExistentialPredicate
U_ = logic.UniversalPredicate


class Evaluator(ExpressionBasicEvaluator, TypedSymbolTableEvaluator):
    pass


def test_build_constants():
    a = C_('ab')
    assert a.type is str and a.value == 'ab'

    b = C_(['a'])
    assert b.type is Sequence[str]
    assert b.value[0].type is str and b.value[0].value == 'a'

    b = C_(('a', 1))
    assert b.type is Tuple[str, int]
    assert b.value[0].type is str and b.value[0].value == 'a'
    assert b.value[1].type is int and b.value[1].value == 1

    b = C_({'a'})
    assert b.type is AbstractSet[str]
    v = b.value.pop()
    assert v.type is str and v.value == 'a'

    b = C_({'a': 1})
    assert b.type is Mapping[str, int]
    assert len(b.value) == 1
    assert C_('a') in b.value
    assert C_(1) in b.value.values()


def test_pickle_constants():
    a = C_('ab')
    assert pickle.loads(pickle.dumps(a)) == a
    b = C_(['a'])
    assert pickle.loads(pickle.dumps(b)) == b
    c = C_(('a', 1))
    assert pickle.loads(pickle.dumps(c)) == c
    d = C_({'a'})
    assert pickle.loads(pickle.dumps(d)) == d
    e = C_({'a': 1})
    assert pickle.loads(pickle.dumps(e)) == e


def test_fresh_symbol():
    s1 = S_.fresh()
    s2 = S_.fresh()
    s3 = S_[int].fresh()
    s4 = S_('a')

    assert isinstance(s1, S_) and s1.type is Unknown
    assert s1 != s2
    assert s1 != s2 and s2 != s3 and s3.type is int
    assert s1.is_fresh
    assert s2.is_fresh
    assert s3.is_fresh
    assert not s4.is_fresh


def test_fresh_symbol_2():
    # Reset sequence generators as if the program just started
    if hasattr(S_, "_fresh_generator_"):
        del S_._fresh_generator_
    if hasattr(S_[str], "_fresh_generator_"):
        del S_[str]._fresh_generator_

    s2 = S_[str].fresh()
    s1 = S_.fresh()

    assert s1.name != s2.name


def evaluate(expression, **kwargs):
    ebe = Evaluator()
    for k, v in kwargs.items():
        if not isinstance(v, Expression):
            v = C_(v)
        ebe.symbol_table[k] = v
    return ebe.walk(expression)


def test_expression_behaves_as_object():
    a = S_('a')

    with pytest.raises(AttributeError):
        c = a.b

    with expressions_behave_as_objects():
        c = a.b

    assert c.functor.value is getattr
    assert c.args[0] is a
    assert c.args[1] == C_('b')


def test_symbol_application():
    a = S_('a')
    b = S_('b')

    oadd = C_[Callable[[int, int], int]](op.add)
    osub = C_[Callable[[int, int], int]](op.sub)
    omul = C_[Callable[[int, int], int]](op.mul)
    c = oadd(C_[int](2), C_[int](3))
    assert c.functor == oadd
    assert all((e1.value == e2 for e1, e2 in zip(c.args, (2, 3))))
    assert evaluate(c) == 5

    fva = oadd(a, C_[int](3))
    fvb = osub(fva, C_[int](10))
    fvc = omul(fvb, b)

    assert a in fva._symbols and (len(fva._symbols) == 1)
    assert evaluate(fva, a=C_(2)) == 5
    assert a in fvb._symbols and (len(fvb._symbols) == 1)
    assert evaluate(fvb, a=C_(2)) == -5
    assert b in fvc._symbols and (len(fvc._symbols) == 2)
    assert isinstance(evaluate(fvc, b=C_(2)), expressions.FunctionApplication)
    assert evaluate(
        evaluate(fvc, b=C_(3)), a=C_(2)
    ) == evaluate(
        fvc, a=C_(2), b=C_(3)
    ) == -15


def test_or_operations():

    fx = F_(S_('f'), (S_('x'),))
    gx = F_(S_('b'), (S_('x'),))

    res = fx | gx
    assert res == F_(C_(op.or_), (fx, gx))

    x = S_('x')
    y = S_('y')
    assert (x | y) == F_(C_(op.or_), (x, y))


def test_symbol_method_and_operator():
    with expressions_behave_as_objects():

        a = S_('a')
        fva = a.__len__()
        fvb = a - C_[int](4)
        fvc = C_[int](4) - a
        fve = a[C_(2)]

    assert evaluate(a, a=C_(1)) == 1
    assert evaluate(fva, a=C_[AbstractSet[int]](set((C_(1),)))) == 1
    assert evaluate(fvb, a=C_(1)) == -3
    assert evaluate(fvc, a=C_(1)) == 3
    assert evaluate(fvc * C_(2), a=C_(1)) == 6
    assert evaluate(a | C_(False), a=True)
    assert evaluate(fve, a=C_([C_(x) for x in (0, 1, 2)])) == 2


def test_constant_method_and_operator():
    with expressions_behave_as_objects():
        a = C_[int](1)
        fva = a + C_(1)
        b = C_[AbstractSet[int]]({C_(1)})
        fvb = b.__len__()
        fbc = b.union(C_({C_(1), C_(2)}))

    assert a == 1
    assert evaluate(a) == 1
    assert evaluate(fva) == 2
    assert evaluate(fvb) == 1
    assert evaluate(fbc).value == {1, 2}


def test_lambda_expression():
    l_ = L_[Callable[[], int]](tuple(), C_[int](2))
    fa = F_[int](l_, tuple())
    assert evaluate(fa) == 2

    x = S_[int]('x')
    l_ = L_[Callable[[int], int]]((x,), C_[int](2) + x)
    fa = F_[int](l_, (C_[int](2),))

    assert evaluate(fa) == 4

    y = S_[int]('y')
    l_ = L_[Callable[[int, int], int]]((x, y), x + y)
    fa = F_[int](l_, (C_[int](2), C_[int](3),))

    assert evaluate(fa) == 5


def test_symbol_wrapping():
    def f(a: int) -> float:
        '''
        test help
        '''
        return 2. * int(a)

    fva = C_(f)
    x = S_[int]('x')
    fvb = fva(x)
    assert fva.__annotations__ == f.__annotations__
    assert fva.__doc__ == f.__doc__
    assert fva.__name__ == f.__name__
    assert fva.__qualname__ == f.__qualname__
    assert fva.__module__ == f.__module__
    assert inspect.signature(fva) == inspect.signature(f)
    assert fvb.type == float
    assert x.type == int


def test_compatibility_for_pattern_matching():
    for symbol_name in dir(expressions):
        symbol = getattr(expressions, symbol_name)
        if not (
            type(symbol) == type and
            issubclass(symbol, expressions.Expression) and
            symbol != expressions.Expression and
            symbol != expressions.Definition
        ):
            continue
        signature = inspect.signature(symbol)
        argnames = [
            name for name, param in signature.parameters.items()
            if param.default == signature.empty
        ]
        args = (..., ) * len(argnames)
        instance = symbol(*args)

        for argname in argnames:
            assert getattr(instance, argname) == ...


def test_equality():
    assert C_(1) == C_(1)
    assert C_(1) != C_(2)
    assert S_('a') == S_('a')
    assert S_('a') != S_('b')
    assert S_('a')(C_(1)) == S_('a')(C_(1))
    assert S_('a')(C_(1)) != S_('b')(C_(1))
    assert S_('a')(C_(1)) != S_('a')(C_(2))

    assert C_((C_(1), C_(2))) == C_((C_(1), C_(2)))
    assert C_((C_(1), C_(2))) != C_((C_(1), C_(3)))

    assert C_((C_(1), C_(2))) != C_((C_(1)))
    assert S_('a')(C_(1)) != S_('a')(C_(1), C_(2))


def test_instance_check():
    c = C_[int](2)
    assert isinstance(c, expressions.Constant)
    assert isinstance(c, expressions.Constant[int])
    assert issubclass(c.__class__, C_)
    assert c.type is int


def test_typecast_check():
    s = S_('a')
    s_float = s.cast(float)

    assert isinstance(s, expressions.Symbol)
    assert s.type is Unknown
    assert isinstance(s_float, expressions.Symbol)
    assert isinstance(s_float, expressions.Symbol[float])
    assert issubclass(s.__class__, S_)
    assert s_float.type is float
    assert s_float.name == 'a'


def test_fa_composition_symbols_correctly_propagated():
    fa1 = S_('fa1')
    fa2 = S_('fa2')
    fa3 = S_('fa3')
    a = S_('a')
    b = S_('b')
    c = S_('c')
    d = S_('d')
    e = S_('e')

    expression = fa1(a, fa2(b, fa3(c, d), e))
    expression_block = expressions.ExpressionBlock((
        fa1(a), fa2(b, fa3(c, d), e)
    ))

    for symbol in [a, b, c, d, e]:
        assert symbol in expression._symbols
        assert symbol in expression_block._symbols


def test_apply_unapply():
    a = C_(1)
    b = a.apply(*a.unapply())
    assert a is not b and a == b

    a = F_(S_('a'), (C_(1),))
    b = a.apply(*a.unapply())
    assert a is not b and a == b


def test_nested_existentials():
    x = S_('x')
    y = S_('y')
    P = S_('P')
    Q = S_('Q')

    exp = E_(x, E_(y, P(x) & Q(y)))

    assert exp._symbols == {Q, P}


def test_nested_universals():
    x = S_('x')
    y = S_('y')
    P = S_('P')
    Q = S_('Q')

    exp = U_(x, U_(y, P(x) & Q(y)))

    assert exp._symbols == {Q, P}


def test_fresh_symbol_subclass():
    class TestSymbol(expressions.Symbol):
        pass

    assert isinstance(TestSymbol.fresh(), TestSymbol)
    assert isinstance(TestSymbol[int].fresh(), TestSymbol[int])


def test_numpy_ufunc_no_qualname_repr():
    exp = C_(numpy.exp)
    repr(exp)


def test_type_printing_option():
    x = S_[int]('x')
    a = C_[str]('a')

    config.enable_expression_type_printing()
    assert repr(x) == "S{x: int}"
    assert repr(a) == "C{'a': str}"

    config.disable_expression_type_printing()
    assert repr(x) == "S{x}"
    assert repr(a) == "C{'a'}"



import operator
from typing import AbstractSet, Tuple

import pytest

from ...datalog.basic_representation import WrappedRelationalAlgebraSet
from ...expression_walker import ExpressionWalker
from ...expressions import Constant, FunctionApplication, Symbol
from ...utils import NamedRelationalAlgebraFrozenSet
from ..optimisers import (
    EliminateTrivialProjections,
    PushInSelections,
    RelationalAlgebraOptimiser,
    RenameOptimizations,
    RewriteSelections,
    SimplifyExtendedProjectionsWithConstants
)
from ..relational_algebra import (
    ColumnInt,
    ColumnStr,
    EquiJoin,
    ExtendedProjection,
    FullOuterNaturalJoin,
    FunctionApplicationListMember,
    GroupByAggregation,
    LeftNaturalJoin,
    NameColumns,
    NaturalJoin,
    Product,
    Projection,
    RenameColumn,
    RenameColumns,
    ReplaceNull,
    Selection,
    eq_,
    str2columnstr_constant
)

C_ = Constant
ne = Constant(operator.ne)


@pytest.fixture
def r1():
    return WrappedRelationalAlgebraSet([(i, i * 2) for i in range(10)])


@pytest.fixture
def r2():
    return WrappedRelationalAlgebraSet([(i * 2, i * 3) for i in range(10)])


@pytest.fixture
def rs():
    return Symbol[AbstractSet]('R')


@pytest.fixture
def str_columns():
    return tuple(str2columnstr_constant(n) for n in ('a', 'b', 'c', 'd', 'e'))


def test_selection_reorder(r1):
    raop = RelationalAlgebraOptimiser()
    s = Selection(C_(r1), eq_(C_(ColumnInt(0)), C_(1)))
    assert raop.walk(s) is s

    s1 = Selection(C_(r1), eq_(C_(1), C_(ColumnInt(0))))
    assert raop.walk(s1) == s

    s = Selection(C_(r1), eq_(C_(ColumnInt(0)), C_(ColumnInt(1))))
    assert raop.walk(s) is s

    s1 = Selection(C_(r1), eq_(C_(ColumnInt(1)), C_(ColumnInt(0))))
    assert raop.walk(s1) == s

    s_in = Selection(C_(r1), eq_(C_(ColumnInt(1)), C_(ColumnInt(1))))
    s_out = Selection(s_in, eq_(C_(ColumnInt(0)), C_(ColumnInt(1))))
    assert raop.walk(s_out) is s_out

    s_in1 = Selection(C_(r1), eq_(C_(ColumnInt(0)), C_(ColumnInt(1))))
    s_out1 = Selection(s_in1, eq_(C_(ColumnInt(1)), C_(ColumnInt(1))))
    assert raop.walk(s_out1) == s_out


def test_push_selection_equijoins(r1, r2):
    raop = RelationalAlgebraOptimiser()
    s2 = Selection(
        EquiJoin(
            C_(r1), (C_(ColumnInt(0)),),
            C_(r2), (C_(ColumnInt(0)),)
        ),
        eq_(C_(ColumnInt(0)), C_(1))
    )
    s2_res = EquiJoin(
        Selection(
            C_(r1),
            eq_(C_(ColumnInt(0)), C_(1))
        ),
        (C_(ColumnInt(0)),),
        C_(r2), (C_(ColumnInt(0)),)
    )

    assert raop.walk(s2) == s2_res

    s2 = Selection(
        EquiJoin(
            C_(r1), (C_(ColumnInt(0)),),
            C_(r2), (C_(ColumnInt(0)),)
        ),
        eq_(C_(ColumnInt(2)), C_(1))
    )
    s2_res = EquiJoin(
        C_(r1),
        (C_(ColumnInt(0)),),
        Selection(
            C_(r2),
            eq_(C_(ColumnInt(0)), C_(1))
        ),
        (C_(ColumnInt(0)),)
    )

    assert raop.walk(s2) == s2_res

    s2 = Selection(
        EquiJoin(
            C_(r1), (C_(ColumnInt(0)),),
            C_(r2), (C_(ColumnInt(0)),)
        ),
        eq_(C_(ColumnInt(0)), C_(ColumnInt(1)))
    )
    s2_res = EquiJoin(
        Selection(
            C_(r1),
            eq_(C_(ColumnInt(0)), C_(ColumnInt(1)))
        ),
        (C_(ColumnInt(0)),),
        C_(r2),
        (C_(ColumnInt(0)),)
    )

    assert raop.walk(s2) == s2_res

    s2 = Selection(
        EquiJoin(
            C_(r1), (C_(ColumnInt(0)),),
            C_(r2), (C_(ColumnInt(0)),)
        ),
        eq_(C_(ColumnInt(2)), C_(ColumnInt(3)))
    )
    s2_res = EquiJoin(
        C_(r1),
        (C_(ColumnInt(0)),),
        Selection(
            C_(r2),
            eq_(C_(ColumnInt(0)), C_(ColumnInt(1)))
        ),
        (C_(ColumnInt(0)),)
    )

    assert raop.walk(s2) == s2_res

    s2 = Selection(
        EquiJoin(
            C_(r1), (C_(ColumnInt(0)),),
            C_(r2), (C_(ColumnInt(0)),)
        ),
        eq_(C_(ColumnInt(1)), C_(ColumnInt(2)))
    )
    assert raop.walk(s2) == s2


def test_push_and_infer_equijoins(r1, r2):
    raop = RelationalAlgebraOptimiser()
    inner = Product((C_(r1), C_(r2)))
    formula1 = eq_(C_(ColumnInt(0)), C_(ColumnInt(1)))
    s = Selection(inner, formula1)
    assert raop.walk(s) == Product((Selection(C_(r1), formula1), C_(r2)))

    inner = Product((C_(r1), C_(r2)))
    formula2 = eq_(C_(ColumnInt(2)), C_(ColumnInt(3)))
    s = Selection(inner, formula2)
    res = raop.walk(s)
    expected_res = Product((C_(r1), Selection(C_(r2), formula1)))
    assert res == expected_res

    inner = Product((C_(r1), C_(r2)))
    formula3 = eq_(C_(ColumnInt(0)), C_(ColumnInt(3)))
    s = Selection(inner, formula3)
    assert raop.walk(s) == EquiJoin(
        C_(r1),
        (C_(ColumnInt(0)),),
        C_(r2),
        (C_(ColumnInt(1)),),
    )

    inner = Product((C_(r1), C_(r2), C_(r1)))
    formula3 = eq_(C_(ColumnInt(0)), C_(ColumnInt(3)))
    s = Selection(inner, formula3)
    assert raop.walk(s) == Product((
        EquiJoin(
            C_(r1),
            (C_(ColumnInt(0)),),
            C_(r2),
            (C_(ColumnInt(1)),),
        ),
        C_(r1)
    ))

    raop = RelationalAlgebraOptimiser()
    inner = Product((C_(r1), C_(r2)))
    formula4 = eq_(C_(ColumnInt(0)), C_(1))
    s = Selection(inner, formula4)
    assert raop.walk(s) == Product(
        (Selection(C_(r1), formula4), C_(r2))
    )

    raop = RelationalAlgebraOptimiser()
    inner = Product((C_(r1), C_(r2)))
    formula5 = eq_(C_(ColumnInt(2)), C_(1))
    s = Selection(inner, formula5)
    res = raop.walk(s)
    theoretical_res = Product(
        (C_(r1), Selection(C_(r2), formula4))
    )
    assert res == theoretical_res


def test_push_in_optimiser():
    class Opt(PushInSelections, ExpressionWalker):
        pass

    opt = Opt()

    r1 = Symbol('r1')
    r2 = Symbol('r2')
    a = Constant[ColumnStr](ColumnStr('a'))
    b = Constant[ColumnStr](ColumnStr('b'))
    c = Constant[ColumnStr](ColumnStr('c'))
    op = Symbol('op')

    exp1 = NaturalJoin(r1, r2)
    assert opt.walk(exp1) is exp1

    formula = op(a)
    exp2 = Selection(NaturalJoin(RenameColumn(r1, b, a), r2), formula)
    res = opt.walk(exp2)
    exp_res = NaturalJoin(Selection(
        RenameColumn(r1, b, a), formula),
        r2
    )
    assert res == exp_res

    exp3 = Selection(NaturalJoin(r2, RenameColumn(r1, b, a)), formula)
    res = opt.walk(exp3)
    exp_res = NaturalJoin(
        r2,
        Selection(
            RenameColumn(r1, b, a), formula
        )
    )
    assert res == exp_res

    formula = op(a)
    exp4 = Selection(LeftNaturalJoin(RenameColumn(r1, b, a), r2), formula)
    res = opt.walk(exp4)
    exp_res = LeftNaturalJoin(Selection(
        RenameColumn(r1, b, a), formula),
        r2
    )
    assert res == exp_res

    exp5 = Selection(LeftNaturalJoin(r2, RenameColumn(r1, b, a)), formula)
    res = opt.walk(exp5)
    exp_res = LeftNaturalJoin(
        r2,
        Selection(
            RenameColumn(r1, b, a), formula
        )
    )
    assert res == exp_res

    exp6 = Selection(
        ExtendedProjection(
                NameColumns(r1, (c,)),
                (
                    FunctionApplicationListMember(c, a),
                    FunctionApplicationListMember(Constant(1), b)
                )
        ),
        op(a)
    )
    res = opt.walk(exp6)
    exp_res = ExtendedProjection(
        Selection(
            NameColumns(r1, (c,)),
            op(c)
        ),
        (
            FunctionApplicationListMember(c, a),
            FunctionApplicationListMember(Constant(1), b)
        )
    )
    assert res == exp_res

    exp7 = Selection(
        ExtendedProjection(
                NameColumns(r1, (c,)),
                (
                    FunctionApplicationListMember(c, a),
                    FunctionApplicationListMember(Constant(1), b)
                )
        ),
        op(b)
    )
    res = opt.walk(exp7)

    assert res == exp7

    exp8 = Selection(
        GroupByAggregation(
                NameColumns(r1, (a, b)),
                (a,),
                (
                    FunctionApplicationListMember(op, b),
                )
        ),
        op(a)
    )
    res = opt.walk(exp8)

    exp_res = GroupByAggregation(
        Selection(NameColumns(r1, (a, b)), op(a)),
        (a,),
        (
            FunctionApplicationListMember(op, b),
        )
    )

    assert res == exp_res

    exp9 = Selection(
        ReplaceNull(
            NameColumns(r1, (a, b, c)),
            a,
            Constant(1.0)
        ),
        op(a)
    )
    res = opt.walk(exp9)
    assert res == exp9

    exp9 = Selection(
        ReplaceNull(
            NameColumns(r1, (a, b, c)),
            b,
            Constant(1.0)
        ),
        op(a)
    )
    res = opt.walk(exp9)
    exp_res = ReplaceNull(
        Selection(
            NameColumns(r1, (a, b, c)),
            op(a)
        ),
        b,
        Constant(1.0)
    )
    assert res == exp_res

    exp9 = Selection(
        Projection(
            NameColumns(r1, (a, b, c)),
            (a, c)
        ),
        op(a)
    )
    res = opt.walk(exp9)
    exp_res = Projection(
        Selection(
            NameColumns(r1, (a, b, c)),
            op(a)
        ),
        (a, c)
    )
    assert res == exp_res


def test_eliminate_trivial_projections_optimiser(r1):
    class Opt(EliminateTrivialProjections, ExpressionWalker):
        pass

    opt = Opt()

    rc = Constant(r1)
    exp = Projection(
        rc,
        (Constant[ColumnInt](ColumnInt(0)), Constant[ColumnInt](ColumnInt(1)))
    )

    res = opt.walk(exp)
    assert res is rc

    a = Constant[ColumnStr](ColumnStr('a'))
    b = Constant[ColumnStr](ColumnStr('b'))
    rv = NamedRelationalAlgebraFrozenSet(
        columns=('a', 'b'),
        iterable=r1
    )

    r = C_[AbstractSet[Tuple[int, int]]](rv)

    exp1 = Projection(r, (a, b))
    res1 = opt.walk(exp1)
    assert res1 is r

    r0 = Symbol('r0')
    exp = Projection(Projection(r0, (a, b)), (a,))
    res = opt.walk(exp)
    assert res == Projection(r0, (a,))

    exp = Projection(Projection(r0, (a, b)), (a, b))
    res = opt.walk(exp)
    assert res == Projection(r0, (a, b))

    exp = ExtendedProjection(
        r0,
        (
            FunctionApplicationListMember(a, a),
            FunctionApplicationListMember(b, b)
        )
    )
    res = opt.walk(exp)
    assert res == Projection(r0, (a, b))

    exp = Projection(exp, (a,))
    res = opt.walk(exp)
    assert res == Projection(r0, (a,))


def test_simple_extended_projection_to_rename(rs, str_columns):
    class Opt(EliminateTrivialProjections, ExpressionWalker):
        pass

    a, b, c, d, _ = str_columns

    exp = ExtendedProjection(
        NameColumns(rs, (a, c)),
        (
            FunctionApplicationListMember(a, b),
            FunctionApplicationListMember(c, d)
        )
    )

    res = Opt().walk(exp)

    assert res == RenameColumns(
        exp.relation,
        (
            (a, b),
            (c, d)
        )
    )

    exp = ExtendedProjection(
        NameColumns(rs, (a, c)),
        (
            FunctionApplicationListMember(a, b),
        )
    )

    res = Opt().walk(exp)

    assert res == res


@pytest.mark.parametrize(
    "join_op",
    [NaturalJoin, LeftNaturalJoin, FullOuterNaturalJoin]
)
def test_composite_extended_projection_join(r1, str_columns, join_op):
    class Opt(SimplifyExtendedProjectionsWithConstants, ExpressionWalker):
        pass

    opt = Opt()

    a, b, c, d, _ = str_columns
    r1 = NameColumns(Constant(r1), (a, b))

    exp = ExtendedProjection(
        join_op(
            ExtendedProjection(
                r1, (
                    FunctionApplicationListMember(Constant(1), c),
                    FunctionApplicationListMember(a, a)
                )
            ),
            r1
        ),
        (
            FunctionApplicationListMember(c + c, d),
            FunctionApplicationListMember(a, a),
            FunctionApplicationListMember(b, b)
        )
    )

    res = opt.walk(exp)
    exp = ExtendedProjection(
        join_op(
            ExtendedProjection(r1, (FunctionApplicationListMember(a, a),)),
            r1),
        (
            FunctionApplicationListMember(
                FunctionApplication(
                    Constant(operator.add),
                    (Constant(1), Constant(1))
                ),
                d
            ),
            FunctionApplicationListMember(a, a),
            FunctionApplicationListMember(b, b)
        )
    )
    assert res == exp

    exp = ExtendedProjection(
        join_op(
            r1,
            ExtendedProjection(
                r1, (
                    FunctionApplicationListMember(Constant(1), c),
                    FunctionApplicationListMember(a, a)
                )
            )
        ),
        (
            FunctionApplicationListMember(c + c, d),
            FunctionApplicationListMember(a, a),
            FunctionApplicationListMember(b, b)
        )
    )

    res = opt.walk(exp)
    exp = ExtendedProjection(
        join_op(
            r1,
            ExtendedProjection(r1, (FunctionApplicationListMember(a, a),))),
        (
            FunctionApplicationListMember(
                FunctionApplication(
                    Constant(operator.add),
                    (Constant(1), Constant(1))
                ),
                d
            ),
            FunctionApplicationListMember(a, a),
            FunctionApplicationListMember(b, b)
        )
    )
    assert res == exp


@pytest.mark.parametrize(
    "join_op",
    [NaturalJoin, LeftNaturalJoin, FullOuterNaturalJoin]
)
def test_composite_extended_projection_constant_join(r1, str_columns, join_op):
    class Opt(SimplifyExtendedProjectionsWithConstants, ExpressionWalker):
        pass

    opt = Opt()

    a, b, c, d, _ = str_columns
    r1 = NameColumns(Constant(r1), (a, b))

    exp = ExtendedProjection(
        join_op(
            ExtendedProjection(
                r1, (
                    FunctionApplicationListMember(Constant(1), c),
                    FunctionApplicationListMember(a, a)
                )
            ),
            RenameColumns(r1, ((b, c),))
        ),
        (
            FunctionApplicationListMember(c + c, d),
            FunctionApplicationListMember(a, a),
            FunctionApplicationListMember(b, b)
        )
    )

    res = opt.walk(exp)
    exp = ExtendedProjection(
        join_op(
            ExtendedProjection(r1, (FunctionApplicationListMember(a, a),)),
            Selection(RenameColumns(r1, ((b, c),)), eq_(c, Constant(1)))
        ),
        (
            FunctionApplicationListMember(
                FunctionApplication(
                    Constant(operator.add),
                    (Constant(1), Constant(1))
                ),
                d
            ),
            FunctionApplicationListMember(a, a),
            FunctionApplicationListMember(b, b)
        )
    )
    assert res == exp


@pytest.mark.parametrize(
    "join_op",
    [NaturalJoin, LeftNaturalJoin, FullOuterNaturalJoin]
)
def test_composite_extended_projection_function_join(r1, str_columns, join_op):
    class Opt(SimplifyExtendedProjectionsWithConstants, ExpressionWalker):
        pass

    opt = Opt()

    a, b, c, d, _ = str_columns
    r1 = NameColumns(Constant(r1), (a, b))

    exp = ExtendedProjection(
        join_op(
            ExtendedProjection(
                r1, (
                    FunctionApplicationListMember(b + Constant(1), c),
                    FunctionApplicationListMember(a, a)
                )
            ),
            r1
        ),
        (
            FunctionApplicationListMember(c + c, d),
            FunctionApplicationListMember(a, a),
            FunctionApplicationListMember(b, b)
        )
    )

    res = opt.walk(exp)
    fresh_column = [
        s for s in res.relation.columns()
        if s.value.startswith('fresh')
    ][0]
    exp = ExtendedProjection(
        join_op(
            ExtendedProjection(
                r1,
                (
                    FunctionApplicationListMember(a, a),
                    FunctionApplicationListMember(b, fresh_column),
                )
            ),
            r1
        ),
        (
            FunctionApplicationListMember(
                FunctionApplication(
                    Constant(operator.add),
                    (fresh_column + Constant(1), fresh_column + Constant(1))
                ),
                d
            ),
            FunctionApplicationListMember(a, a),
            FunctionApplicationListMember(b, b)
        )
    )
    assert res == exp


def test_composite_extended_projection_replacenull(r1, str_columns):
    class Opt(SimplifyExtendedProjectionsWithConstants, ExpressionWalker):
        pass

    opt = Opt()

    a, b, c, _, _ = str_columns
    r1 = NameColumns(Constant(r1), (a, b))

    exp = ReplaceNull(
        ExtendedProjection(
            r1, (
                FunctionApplicationListMember(Constant(1), c),
                FunctionApplicationListMember(a, b)
            )
        ),
        b,
        Constant(1)
    )

    res = opt.walk(exp)
    exp = ExtendedProjection(
        ReplaceNull(
            r1,
            a,
            Constant(1)
        ),
        (
            FunctionApplicationListMember(Constant(1), c),
            FunctionApplicationListMember(a, b)
        )
    )
    assert res == exp


def test_extended_projection_groupby_trivial(r1, str_columns):
    class Opt(SimplifyExtendedProjectionsWithConstants, ExpressionWalker):
        pass

    opt = Opt()

    a, b, c, d, _ = str_columns
    r1 = NameColumns(Constant(r1), (a, b))

    exp = GroupByAggregation(
        ExtendedProjection(
            r1, (
                FunctionApplicationListMember(Constant(1), c),
                FunctionApplicationListMember(a, b)
            )
        ),
        (b,),
        (FunctionApplicationListMember(Constant(sum)(c), d),)
    )

    res = opt.walk(exp)
    exp = GroupByAggregation(
        ExtendedProjection(
            r1, (
                FunctionApplicationListMember(a, b),
            )
        ),
        (b,),
        (FunctionApplicationListMember(Constant(len)(), d),)
    )
    assert res == exp


def test_rename_column_to_rename_columns(rs, str_columns):
    class Opt(RenameOptimizations, ExpressionWalker):
        pass

    opt = Opt()

    exp = RenameColumn(rs, str_columns[0], str_columns[1])

    res = opt.walk(exp)

    assert res == RenameColumns(rs, ((str_columns[0], str_columns[1]),))


def test_trivial_rename_columns(rs, str_columns):
    class Opt(RenameOptimizations, ExpressionWalker):
        pass

    opt = Opt()

    exp = RenameColumns(rs, tuple())

    res = opt.walk(exp)

    assert res == rs

    exp = RenameColumns(rs, tuple((s, s) for s in str_columns))

    res = opt.walk(exp)

    assert res == rs


def test_nested_rename_columns(rs, str_columns):
    class Opt(RenameOptimizations, ExpressionWalker):
        pass

    a, b, c, d, e = str_columns

    opt = Opt()

    exp = RenameColumns(
        RenameColumns(rs, ((a, b), (d, e))),
        ((b, c),)
    )

    res = opt.walk(exp)

    assert res == RenameColumns(rs, ((a, c), (d, e)))


def test_nested_rename_columns_extended_projection(rs, str_columns):
    class Opt(RenameOptimizations, ExpressionWalker):
        pass

    a, b, c, d, _ = str_columns

    opt = Opt()

    exp = RenameColumns(
        ExtendedProjection(
            rs,
            (
                FunctionApplicationListMember(a, b),
                FunctionApplicationListMember(Constant(1), c)
            )
        ),
        ((b, c), (c, d))
    )

    res = opt.walk(exp)

    assert res == ExtendedProjection(
        rs,
        (
            FunctionApplicationListMember(a, c),
            FunctionApplicationListMember(Constant(1), d)
        )
    )


def test_nested_rename_columns_groupby_agg(rs, str_columns):
    class Opt(RenameOptimizations, ExpressionWalker):
        pass

    a, b, c, _, e = str_columns

    opt = Opt()

    exp = RenameColumns(
        GroupByAggregation(
            rs,
            (a,),
            (
                FunctionApplicationListMember(Constant(sum)(c), b),
            )
        ),
        ((a, e),)
    )

    res = opt.walk(exp)

    assert res == GroupByAggregation(
        RenameColumns(rs, ((a, e),)),
        (e,),
        (
            FunctionApplicationListMember(Constant(sum)(c), b),
        )
    )

    exp = RenameColumns(
        GroupByAggregation(
            rs,
            (a,),
            (FunctionApplicationListMember(Constant(sum)(c), b),)
        ),
        ((a, b), (b, c))
    )

    res = opt.walk(exp)

    assert res == GroupByAggregation(
        RenameColumns(rs, ((a, b),)),
        (b,),
        (
            FunctionApplicationListMember(Constant(sum)(c), c),
        )
    )

    exp = RenameColumns(
        GroupByAggregation(
            rs,
            (a,),
            (FunctionApplicationListMember(Constant(sum)(b), b),)
        ),
        ((a, b), (b, c))
    )

    res = opt.walk(exp)

    assert res == RenameColumns(
        GroupByAggregation(
            rs,
            (a,),
            (FunctionApplicationListMember(Constant(sum)(b), c),)
        ),
        ((a, b),)
    )


def test_rename_name(rs, str_columns):
    class Opt(RenameOptimizations, ExpressionWalker):
        pass

    a, b, c, _, _ = str_columns

    opt = Opt()

    exp = RenameColumns(NameColumns(rs, (a, b)), ((a, c), (b, a)))

    res = opt.walk(exp)

    assert res == NameColumns(rs, ((c, a)))


def test_rename_selection(rs, str_columns):
    class Opt(RenameOptimizations, ExpressionWalker):
        pass

    a, _, c, _, _ = str_columns

    opt = Opt()

    one = Constant(1)
    two = Constant(2)

    exp = RenameColumns(Selection(rs, eq_(a + two, one)), ((a, c),))

    res = opt.walk(exp)

    assert res == Selection(RenameColumns(rs, ((a, c),)), eq_(c + two, one))


def test_rename_projection(rs, str_columns):
    class Opt(RenameOptimizations, ExpressionWalker):
        pass

    a, b, c, _, _ = str_columns

    opt = Opt()

    exp = RenameColumns(Projection(rs, (a, b)), ((a, c),))

    res = opt.walk(exp)

    assert res == Projection(RenameColumns(rs, ((a, c),)), (c, b))


def test_rename_replacenull(rs, str_columns):
    class Opt(RenameOptimizations, ExpressionWalker):
        pass

    a, _, c, _, _ = str_columns

    opt = Opt()

    zero = Constant(0)

    exp = RenameColumns(ReplaceNull(rs, a, zero), ((a, c),))

    res = opt.walk(exp)

    assert res == ReplaceNull(RenameColumns(rs, ((a, c),)), c, zero)


def test_nested_selections(rs, str_columns):
    class Opt(RewriteSelections, ExpressionWalker):
        pass

    a, b, _, _, _ = str_columns

    opt = Opt()

    and_ = Constant(operator.and_)
    one = Constant(1)

    exp = Selection(Selection(rs, eq_(a, one)), eq_(b + one, one))

    res = opt.walk(exp)

    assert res == Selection(rs, and_(eq_(a, one), eq_(b + one, one)))


def test_push_unnamed_selections_up(r1):
    raop = RelationalAlgebraOptimiser()
    x = str2columnstr_constant("x")
    y = str2columnstr_constant("y")

    s = NameColumns(
        Selection(C_(r1), eq_(C_(ColumnInt(0)), C_(ColumnInt(1)))),
        (x, y)
    )
    res = Selection(NameColumns(C_(r1), (x, y)), eq_(x, y))
    assert raop.walk(s) == res

    s = NameColumns(
        Selection(C_(r1), eq_(C_(ColumnInt(1)), C_(ColumnInt(0)))),
        (x, y)
    )
    res = Selection(NameColumns(C_(r1), (x, y)), eq_(y, x))
    assert raop.walk(s) == res

    s = Projection(
        Selection(C_(r1), eq_(C_(ColumnInt(1)), C_(0))),
        (C_(ColumnInt(1)),)
    )
    res = Selection(
        Projection(C_(r1), (C_(ColumnInt(1)),)), eq_(C_(ColumnInt(1)), C_(0))
    )
    assert raop.walk(s) == res

    s = Projection(
        Selection(C_(r1), eq_(C_(ColumnInt(1)), C_(ColumnInt(0)))),
        (C_(ColumnInt(0)),)
    )
    res = Projection(
        Selection(C_(r1), eq_(C_(ColumnInt(0)), C_(ColumnInt(1)))),
        (C_(ColumnInt(0)),)
    )
    assert raop.walk(s) == res

    s = Projection(
        Selection(C_(r1), eq_(C_(ColumnInt(0)), C_(ColumnInt(1)))),
        (C_(ColumnInt(1)),)
    )
    assert raop.walk(s) == res

    s = Projection(
        Selection(C_(r1), eq_(C_(ColumnInt(1)), C_(ColumnInt(0)))),
        (C_(ColumnInt(0)), C_(ColumnInt(1)))
    )

    res = Selection(
        Projection(C_(r1), (C_(ColumnInt(0)), C_(ColumnInt(1)))),
        eq_(C_(ColumnInt(1)), C_(ColumnInt(0)))
    )

    assert raop.walk(s) == res

    s = Selection(
        Selection(C_(r1), ne(C_(ColumnInt(1)), C_(ColumnInt(2)))),
        eq_(C_(ColumnInt(0)), C_(ColumnInt(1)))
    )

    res = Selection(
        Selection(C_(r1), eq_(C_(ColumnInt(0)), C_(ColumnInt(1)))),
        ne(C_(ColumnInt(0)), C_(ColumnInt(2)))
    )

    assert raop.walk(s) == res



"""
Pytest configuration file.

NOTE: Do not remove the unused dask_sql import !

The dask_sql library uses jpype to start a JVM and access Java objects
from python. This JVM needs to be started before we import neurolang,
otherwise it can cause major side-effects which are hard to track. See
https://github.com/jpype-project/jpype/issues/933 for reference.
"""
import warnings
import pytest
try:
    import dask_sql
    from neurolang.utils.relational_algebra_set.dask_helpers import (
        DaskContextManager,
    )

    HAS_DASK = True
except ImportError:
    HAS_DASK = False
from neurolang import config


def pytest_addoption(parser):
    parser.addoption(
        "--runslow", action="store_true", default=False, help="run slow tests"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


def pytest_sessionstart(session: pytest.Session):
    """
    Hook called after the pytest Session object has been created and
    before performing collection and entering the run test loop.

    The dask-sql library uses the jpype library which starts a JVM and allows
    us to use Java classes from Python. But the JVM will trigger a
    segmentation fault when starting and when interrupting threads and Pythons
    fault handler can intercept these operations and interpret these as
    real faults. So we need to disable faulthandlers which pytest starts
    otherwise we get segmentation faults when running the tests.
    See (https://jpype.readthedocs.io/en/latest/userguide.html#errors-reported-by-python-fault-handler)
    """
    try:
        import faulthandler

        faulthandler.enable()
        faulthandler.disable()
    except ImportError:
        warnings.warn("Faulthandler lib not available.")


@pytest.fixture(autouse=config["RAS"].get("backend", "pandas") == "dask")
def clear_dask_context_after_test_module():
    """
    We use only one DaskContextManager for the application and its context gets
    clustered with objects quite fast when running the tests, so this fixture clears
    the context after each test function.
    """
    yield 0

    DaskContextManager._context = None



from operator import add, eq, lt, mul, pow, sub, truediv

import pytest


from neurolang.logic import ExistentialPredicate

from ....datalog import Conjunction, Fact, Implication, Negation, Union
from ....expressions import (
    Command,
    Constant,
    FunctionApplication,
    Lambda,
    Query,
    Statement,
    Symbol
)
from ....exceptions import UnexpectedTokenError
from ....probabilistic.expressions import (
    PROB,
    Condition,
    ProbabilisticFact
)
from ..standard_syntax import ExternalSymbol, parser


def test_facts():
    res = parser('A(3)')
    assert res == Union((Fact(Symbol('A')(Constant(3))),))

    res = parser('A("x")')
    assert res == Union((Fact(Symbol('A')(Constant('x'))),))

    res = parser("A('x', 3)")
    assert res == Union((Fact(Symbol('A')(Constant('x'), Constant(3))),))

    res = parser(
        'A("x", 3)\n'
        '`http://uri#test-fact`("x")\n'
        'ans():-A(x, y)'
    )
    assert res == Union((
        Fact(Symbol('A')(Constant('x'), Constant(3))),
        Fact(Symbol('http://uri#test-fact')(Constant('x'))),
        Query(
            Symbol('ans')(),
            Conjunction((
                Symbol('A')(Symbol('x'), Symbol('y')),
            ))
        )
    ))


def test_rules():
    A = Symbol('A')
    B = Symbol('B')
    C = Symbol('C')
    f = Symbol('f')
    x = Symbol('x')
    y = Symbol('y')
    z = Symbol('z')
    res = parser('A(x):-B(x, y), C(3, z)')
    assert res == Union((
        Implication(A(x), Conjunction((B(x, y), C(Constant(3), z)))),
    ))

    res = parser('A(x):-~B(x)')
    assert res == Union((
        Implication(A(x), Conjunction((Negation(B(x)),))),
    ))

    res = parser('A(x):-B(x, ...)')
    fresh_arg = res.formulas[0].antecedent.formulas[0].args[1]
    assert isinstance(fresh_arg, Symbol)
    assert res == Union((
        Implication(A(x), Conjunction((B(x, fresh_arg),))),
    ))

    res = parser('A(x):-B(x, y), C(3, z), (z == 4)')
    assert res == Union((
        Implication(
            A(x),
            Conjunction((
                B(x, y), C(Constant(3), z), Constant(eq)(z, Constant(4))
            ))
        ),
    ))

    res = parser('A(x):-B(x + 5 * 2, y), C(3, z), (z == 4)')
    assert res == Union((
        Implication(
            A(x),
            Conjunction((
                B(
                    Constant(add)(
                        x,
                        Constant(mul)(Constant(5), Constant(2))),
                    y
                ),
                C(Constant(3), z), Constant(eq)(z, Constant(4))
            ))
        ),
    ))

    res = parser('A(x):-B(x / 2, y), C(3, z), (z == 4)')
    assert res == Union((
        Implication(
            A(x),
            Conjunction((
                B(
                    Constant(truediv)(x, Constant(2)),
                    y
                ),
                C(Constant(3), z), Constant(eq)(z, Constant(4))
            ))
        ),
    ))

    res = parser('A(x):-B(f(x), y), C(3, z), (z == 4)')
    assert res == Union((
        Implication(
            A(x),
            Conjunction((
                B(f(x), y),
                C(Constant(3), z), Constant(eq)(z, Constant(4))
            ))
        ),
    ))

    res = parser('A(x):-B(x + (-5), "a")')
    assert res == Union((
        Implication(
            A(x),
            Conjunction((
                B(
                    Constant(add)(x, Constant(-5)),
                    Constant("a")
                ),
            ))
        ),
    ))

    res = parser('A(x):-B(x - 5 * 2, @y ** -2)')
    assert res == Union((
        Implication(
            A(x),
            Conjunction((
                B(
                    Constant(sub)(
                        x,
                        Constant(mul)(Constant(5), Constant(2))
                    ),
                    Constant(pow)(ExternalSymbol('y'), Constant(-2))
                ),
            ))
        ),
    ))


def test_aggregation():
    A = Symbol('A')
    B = Symbol('B')
    f = Symbol('f')
    x = Symbol('x')
    y = Symbol('y')
    res = parser('A(x, f(y)):-B(x, y)')
    expected_result = Union((
        Implication(
            A(x, FunctionApplication(f, (y,))),
            Conjunction((B(x, y),))
        ),
    ))
    assert res == expected_result


def test_uri():
    from rdflib import RDFS

    label = Symbol(name=str(RDFS.label))
    regional_part = Symbol(
        name='http://sig.biostr.washington.edu/fma3.0#regional_part_of'
    )
    x = Symbol('x')
    y = Symbol('y')

    res = parser(f'`{str(label.name)}`(x):-`{str(regional_part.name)}`(x, y)')
    expected_result = Union((
        Implication(
            label(x),
            Conjunction((regional_part(x, y),))
        ),
    ))

    assert res == expected_result


def test_probabilistic_fact():
    A = Symbol('A')
    p = Symbol('p')
    res = parser('p::A(3)')
    assert res == Union((
        Implication(
            ProbabilisticFact(p, A(Constant(3.))),
            Constant(True)
        ),
    ))

    res = parser('0.8::A("a b", 3)')
    assert res == Union((
        Implication(
            ProbabilisticFact(
                Constant(0.8),
                A(Constant("a b"), Constant(3.))
            ),
            Constant(True)
        ),
    ))

    exp = Symbol("exp")
    d = Symbol("d")
    x = Symbol("x")
    B = Symbol("B")
    res = parser("B(x) :: exp(-d / 5.0) :- A(x, d) , (d < 0.8)")
    expected = Union(
        (
            Implication(
                ProbabilisticFact(
                    FunctionApplication(
                        exp,
                        (
                            Constant(truediv)(
                                Constant(mul)(Constant(-1), d), Constant(5.0)
                            ),
                        ),
                    ),
                    B(x),
                ),
                Conjunction((A(x, d), Constant(lt)(d, Constant(0.8)))),
            ),
        )
    )
    assert res == expected


def test_condition():
    A = Symbol('A')
    B = Symbol('B')
    C = Symbol('C')
    x = Symbol('x')
    res = parser('C(x) :- A(x) // B(x)')

    expected = Union((
        Implication(
            C(x),
            Condition(A(x), B(x))
        ),
    ))

    assert res == expected

    res = parser('C(x) :- (A(x), B(x)) // B(x)')

    expected = Union((
        Implication(
            C(x),
            Condition(Conjunction((A(x), B(x))), B(x))
        ),
    ))

    assert res == expected

    res = parser('C(x) :- A(x) // (A(x), B(x))')

    expected = Union((
        Implication(
            C(x),
            Condition(A(x), Conjunction((A(x), B(x))))
        ),
    ))

    assert res == expected


def test_existential():
    A = Symbol("A")
    B = Symbol("B")
    C = Symbol("C")
    x = Symbol("x")
    s1 = Symbol("s1")
    s2 = Symbol("s2")

    res = parser("C(x) :- B(x), exists(s1; A(s1))")
    expected = Union(
        (
            Implication(
                C(x), Conjunction((B(x), ExistentialPredicate(s1, A(s1))))
            ),
        )
    )
    assert res == expected

    res = parser("C(x) :- B(x), ∃(s1 st A(s1))")
    assert res == expected

    res = parser("C(x) :- B(x), exists(s1, s2; A(s1), A(s2))")

    expected = Union(
        (
            Implication(
                C(x),
                Conjunction(
                    (
                        B(x),
                        ExistentialPredicate(
                            s2,
                            ExistentialPredicate(
                                s1, Conjunction((A(s1), A(s2)))
                            ),
                        ),
                    )
                ),
            ),
        )
    )

    assert res == expected

    # try :
    #     res = parser("C(x) :- B(x), exists(s1; )")
    # except UnexpectedToken as e:
    #     raise UnexpectedTokenError from e
    with pytest.raises(UnexpectedTokenError):
        res = parser("C(x) :- B(x), exists(s1; )")

def test_query():
    ans = Symbol("ans")
    B = Symbol("B")
    C = Symbol("C")
    x = Symbol("x")
    y = Symbol("y")
    res = parser("ans(x) :- B(x, y), C(3, y)")
    assert res == Union(
        (Query(ans(x), Conjunction((B(x, y), C(Constant(3), y)))),)
    )


def test_prob_implicit():
    B = Symbol("B")
    C = Symbol("C")
    x = Symbol("x")
    y = Symbol("y")
    res = parser("B(x, PROB, y) :- C(x, y)")
    assert res == Union(
        (Implication(B(x, PROB(x, y), y), Conjunction((C(x, y),))),)
    )


def test_prob_explicit():
    B = Symbol("B")
    C = Symbol("C")
    x = Symbol("x")
    y = Symbol("y")
    res = parser("B(x, PROB(x, y), y) :- C(x, y)")
    assert res == Union(
        (Implication(B(x, PROB(x, y), y), Conjunction((C(x, y),))),)
    )


def test_lambda_definition():
    c = Symbol("c")
    x = Symbol("x")

    res = parser("c := lambda x: x + 1")
    expression = Lambda((x,), FunctionApplication(
        Constant(add), (x, Constant[int](1))
    ))
    expected = Union((
        Statement(c, expression),
    ))
    assert expected == res


def test_lambda_definition_statement():
    c = Symbol("c")
    x = Symbol("x")
    y = Symbol("y")

    res = parser("c(x, y) := x + y")
    expression = Lambda((x, y), FunctionApplication(
        Constant(add), (x, y)
    ))
    expected = Union((
        Statement(c, expression),
    ))
    assert expected == res


def test_lambda_application():
    c = Symbol("c")
    x = Symbol("x")

    res = parser("c := (lambda x: x + 1)(2)")
    expression = FunctionApplication(
        Lambda((x,), FunctionApplication(
            Constant(add), (x, Constant[int](1))
        )),
        (Constant[int](2),)
    )
    expected = Union((
        Statement(c, expression),
    ))
    assert expected == res


def test_command_syntax():
    res = parser('.load_csv(A, "http://myweb/file.csv", B)')
    expected = Union(
        (
            Command(
                Symbol("load_csv"),
                (Symbol("A"), Constant("http://myweb/file.csv"), Symbol("B")),
                (),
            ),
        )
    )
    assert res == expected

    res = parser('.load_csv("http://myweb/file.csv")')
    expected = Union(
        (
            Command(
                Symbol("load_csv"), (Constant("http://myweb/file.csv"),), ()
            ),
        )
    )
    assert res == expected

    res = parser(".load_csv()")
    expected = Union((Command(Symbol("load_csv"), (), ()),))
    assert res == expected

    res = parser('.load_csv(sep=",")')
    expected = Union(
        (Command("load_csv", (), ((Symbol("sep"), Constant(",")),)),)
    )
    assert res == expected

    res = parser('.load_csv(sep=",", header=None, index_col=0)')
    expected = Union(
        (
            Command(
                "load_csv",
                (),
                (
                    (Symbol("sep"), Constant(",")),
                    (Symbol("header"), Symbol("None")),
                    (Symbol("index_col"), Constant(0)),
                ),
            ),
        )
    )
    assert res == expected

    res = parser('.load_csv(A, "http://myweb/file.csv", sep=",", header=None)')
    expected = Union(
        (
            Command(
                "load_csv",
                (Symbol("A"), "http://myweb/file.csv"),
                (
                    (Symbol("sep"), Constant(",")),
                    (Symbol("header"), Symbol("None")),
                ),
            ),
        )
    )
    assert res == expected



import collections
import itertools
import operator

from ..expression_pattern_matching import add_match
from ..expression_walker import (
    ExpressionWalker,
    ReplaceExpressionWalker,
    expression_iterator
)
from ..expressions import Constant, FunctionApplication, Symbol
from ..logic import Conjunction, Disjunction, ExistentialPredicate
from ..logic.expression_processing import (
    extract_logic_atoms,
    extract_logic_free_variables
)
from ..logic.transformations import (
    LogicExpressionWalker,
    RemoveTrivialOperations,
    convert_to_pnf_with_dnf_matrix
)
from ..relational_algebra import (
    ColumnStr,
    Projection,
    Selection,
    int2columnint_constant,
    str2columnstr_constant
)
from ..utils import powerset
from .expression_processing import UnifyVariableEqualities
from .probabilistic_ra_utils import (
    ProbabilisticChoiceSet,
    ProbabilisticFactSet
)

EQ = Constant(operator.eq)
NE = Constant(operator.ne)
RTO = RemoveTrivialOperations()


class HeadVar(ColumnStr):
    pass


def str2columnstr_headvar_constant(name):
    return Constant[HeadVar](
        name, auto_infer_type=False, verify_type=False,
    )


def atom_to_constant_to_RA_conditions(
    atom: FunctionApplication, head_vars: set
) -> set:
    conditions = set()
    for i, arg in enumerate(atom.args):
        if (
            isinstance(arg, Constant) or
            (arg in head_vars)
        ):
            if isinstance(arg, Symbol):
                arg = str2columnstr_headvar_constant(arg.name)
            conditions.add(EQ(int2columnint_constant(i), arg))
    return conditions


class NormalizeNotEquals(LogicExpressionWalker):
    def __init__(self, head_vars):
        self.head_vars = head_vars

    @add_match(FunctionApplication(NE, (Constant, Symbol)))
    def flip_ne_arguments(self, expression):
        return expression.apply(
            NE,
            expression.args[::-1]
        )

    @add_match(FunctionApplication(NE, (Symbol, Symbol)))
    def flip_ne_symbol_arguments(self, expression):
        if (
            expression.args[0] in self.head_vars and
            expression.args[1] not in self.head_vars
        ):
            expression = expression.apply(
                NE,
                expression.args[::-1]
            )

        return expression


def conditions_per_symbol(ucq_query):
    ucq_query_dnf = convert_to_pnf_with_dnf_matrix(ucq_query)
    head_vars = extract_logic_free_variables(ucq_query_dnf)
    ucq_query_dnf = NormalizeNotEquals(head_vars).walk(ucq_query_dnf)

    conditions_per_symbol = collections.defaultdict(set)
    number_of_args = dict()
    for atom in extract_logic_atoms(ucq_query_dnf):
        if not isinstance(atom.functor, Symbol):
            continue
        condition = atom_to_constant_to_RA_conditions(atom, head_vars)
        conditions_per_symbol[atom.functor] |= condition
        number_of_args[atom.functor] = len(atom.args)

    conjunctions = (
        expression for _, expression in expression_iterator(ucq_query_dnf)
        if isinstance(expression, Conjunction) and
        any(
            _formula_is_ne(formula)
            for formula in expression.formulas
        )
    )

    for conjunction in conjunctions:
        nes = (
            formula for formula in conjunction.formulas
            if _formula_is_ne(formula)
        )
        for ne in nes:
            ne_symbol = ne.args[0]
            ne_constant = ne.args[1]
            if isinstance(ne_constant, Symbol):
                if ne_constant in head_vars:
                    ne_constant = str2columnstr_headvar_constant(
                        ne_constant.name
                    )
                else:
                    ne_constant = str2columnstr_constant(ne_constant.name)
            for atom in extract_logic_atoms(conjunction):
                if not isinstance(atom.functor, Symbol):
                    continue
                ne_conditions = {
                    NE(int2columnint_constant(i), ne_constant)
                    for i, s in enumerate(atom.args)
                    if s == ne_symbol
                }
                if ne_conditions:
                    conditions_per_symbol[atom.functor] |= ne_conditions
                    number_of_args[atom.functor] = len(atom.args)

    return conditions_per_symbol, number_of_args


def _formula_is_ne(formula):
    return (
        isinstance(formula, FunctionApplication) and
        formula.functor == NE and
        isinstance(formula.args[0], Symbol)
    )


def _return_value_or_name(v):
    if isinstance(v, Constant):
        return v.value
    else:
        return v.name


def sets_per_symbol(ucq_query):
    cps, number_of_args = conditions_per_symbol(ucq_query)

    complementary = {EQ: NE, NE: EQ}
    symbol_sets = dict()

    for symbol, conditions in cps.items():
        if len(conditions) == 0:
            continue
        column_conditions = collections.defaultdict(set)
        for condition in conditions:
            if condition.functor == NE:
                condition = condition.apply(EQ, condition.unapply()[1])
            column_conditions[condition.args[0]].add((condition,))

        for column, conditions in column_conditions.items():
            column_conditions[column].add(
                tuple(
                    complementary[c[0].functor](*c[0].args) for c in conditions

                )
            )

        set_formulas = list(
            Conjunction(tuple(
                sorted(
                    itertools.chain(*conditions),
                    key=lambda exp: (
                        _return_value_or_name(exp.args[0]),
                        _return_value_or_name(exp.args[1])
                    )
                )
            ))
            for conditions in
            itertools.product(*column_conditions.values())
        )
        conditions_columns = [
            set(
                condition_item.args[0].value
                for condition_item in condition.formulas
                if condition_item.functor == EQ
            )
            for condition in set_formulas
        ]
        conditions_projected_args = [
            tuple(
                int2columnint_constant(i)
                for i in range(number_of_args[symbol])
                # if i not in condition_columns
            )
            for condition_columns in conditions_columns
        ]

        symbol_sets[symbol] = set(
            Projection(Selection(symbol, RTO.walk(condition)), projected_args)
            for condition, projected_args
            in zip(set_formulas, conditions_projected_args)
        )

    return symbol_sets


class ShatterEqualities(ExpressionWalker):
    def __init__(self, symbol_sets, symbol_table, head_vars):
        self.head_vars = head_vars
        self.symbol_sets = symbol_sets
        self.symbol_table = symbol_table
        self.shatter_symbols = dict()

    @add_match(FunctionApplication(Symbol, ...))
    def shatter_symbol(self, expression):
        expression = self._shatter_atom(expression)
        return expression

    def _shatter_atom(self, expression, nes=None):
        if expression.functor not in self.symbol_sets:
            return expression

        sets = self.symbol_sets[expression.functor]
        sets_to_keep = self._identify_sets_to_keep(expression, sets, nes=nes)

        expression, e_vars = self._compute_expression_arguments(
            expression, sets_to_keep
        )

        if len(expression) == 1:
            expression = expression[0]
        else:
            expression = Disjunction(expression)

        for head_var in (
            self.head_vars &
            extract_logic_free_variables(expression)
        ):
            new_var = Symbol.fresh()
            expression = ReplaceExpressionWalker(
                {head_var: new_var}
            ).walk(expression)
            e_vars.add(new_var)

        for e_var in e_vars:
            expression = ExistentialPredicate(e_var, expression)
        return expression

    @add_match(
        Conjunction,
        lambda expression: (
            any(
                _formula_is_ne(formula)
                for formula in expression.formulas
            )
        )
    )
    def shatter_inequalities(self, expression):
        nes = set()
        formulas_to_keep = []
        for formula in expression.formulas:
            if (
                _formula_is_ne(formula) and
                (
                    isinstance(formula.args[1], Constant) or
                    formula.args[1] in self.head_vars
                )
            ):
                if isinstance(formula.args[1], Symbol):
                    functor, args = formula.unapply()
                    args = (
                        args[0],
                        str2columnstr_headvar_constant(args[1].name)
                    )
                    formula = formula.apply(functor, args)
                nes.add(formula)
            else:
                formulas_to_keep.append(formula)

        resulting_formulas = tuple()
        for formula in formulas_to_keep:
            if formula.functor not in self.symbol_sets:
                resulting_formulas += (formula,)
            else:
                resulting_formulas += (self._shatter_atom(formula, nes=nes),)

        return Conjunction(resulting_formulas)

    def _compute_expression_arguments(self, expression, sets_to_keep):
        args = []
        for i, arg in enumerate(expression.args):
            if isinstance(arg, Symbol):
                args.append((i, arg, False))
            elif isinstance(arg, Constant):
                args.append((i, Symbol.fresh(), True))
        formulas = tuple()
        e_vars = set()
        for set_ in sets_to_keep:
            if set_ not in self.shatter_symbols:
                sym = Symbol.fresh()
                self.shatter_symbols[set_] = sym
                self.symbol_table[sym] = set_
            formula_args = []
            for i, arg, is_e_var in args:
                if int2columnint_constant(i) in set_.columns():
                    formula_args.append(arg)
                    if is_e_var:
                        e_vars.add(arg)
            formulas += (self.shatter_symbols[set_](*formula_args),)
        return formulas, e_vars

    def _identify_sets_to_keep(self, expression, sets, nes=None):
        conditions = atom_to_constant_to_RA_conditions(
            expression, self.head_vars
        )
        negative_conditions = set()
        if nes is not None:
            args = list(expression.args)
            for ne in nes:
                if ne.args[0] in expression.args:
                    negative_conditions.add(
                        EQ(
                            int2columnint_constant(args.index(ne.args[0])),
                            ne.args[1]
                        )
                    )
        sets_to_keep = []
        for set_ in sets:
            set_conditions = self._selection_formulas(set_)
            if (
                conditions <= set_conditions and
                set_conditions.isdisjoint(negative_conditions)
            ):
                sets_to_keep.append(set_)
        return sets_to_keep

    def _selection_formulas(self, set_):
        formula = set_.relation.formula
        if isinstance(formula, Conjunction):
            set_conditions = set(formula.formulas)
        else:
            set_conditions = {formula}
        return set_conditions


class RelationalAlgebraSelectionConjunction(ExpressionWalker):
    @add_match(Selection(..., Conjunction))
    def selection_conjunction(self, expression):
        and_ = Constant(operator.and_)
        formulas = expression.formula.formulas
        new_formula = formulas[0]
        for f in formulas[1:]:
            new_formula = and_(new_formula, f)
        return Selection(
            expression.relation,
            new_formula
        )


def extract_variable_support(query, include_position=False):
    support = collections.defaultdict(lambda: set())
    for predicate in extract_logic_atoms(query):
        for i, arg in enumerate(predicate.args):
            if not isinstance(arg, Symbol):
                continue
            if include_position:
                p = (predicate.functor, i)
            else:
                p = predicate.functor
            support[arg].add(p)
    return support


def compute_var_equivalence_groups(query, head_vars):
    head_var_support = {
        k: v
        for k, v in extract_variable_support(query, True).items()
        if k in head_vars
    }

    equivalence_groups = {
        frozenset((v,)): head_var_support[v]
        for v in head_vars
    }
    change = True
    while change:
        change = False
        for g1, g2 in itertools.product(
            equivalence_groups, equivalence_groups
        ):
            if g1 == g2:
                continue
            if equivalence_groups[g1] & equivalence_groups[g2]:
                equivalence_groups[g1 | g2] = (
                    equivalence_groups[g1] |
                    equivalence_groups[g2]
                )
                del equivalence_groups[g1]
                del equivalence_groups[g2]
                change = True
                break

    return set(equivalence_groups.keys())


def contemplate_for_head_var_equivalence(query, head_vars):
    equivalence_groups = compute_var_equivalence_groups(query, head_vars)
    scenarios = []
    for group in equivalence_groups:
        conditions = []
        for subgroup in powerset(group):
            condition = []
            complement = group - set(subgroup)
            for v1, v2 in itertools.product(complement, complement):
                if v1.name < v2.name:
                    condition.append(NE(v1, v2))
            for v1, v2 in itertools.product(subgroup, subgroup):
                if v1.name < v2.name:
                    condition.append(EQ(v1, v2))
            for v1, v2 in itertools.product(
                itertools.islice(subgroup, 1), complement
            ):
                if v1.name < v2.name:
                    condition.append(NE(v1, v2))
            if condition:
                condition = Conjunction(tuple(condition))
                conditions.append(condition)
        scenarios.append(set(conditions))

    scenarios = [
        RTO.walk(Conjunction(tuple(s for s in scenes)))
        for scenes in itertools.product(*scenarios)
    ]

    if len(scenarios) > 0:
        query = Disjunction(tuple(
            RTO.walk(Conjunction((query, scenario)))
            for scenario in scenarios
        ))
        query = UnifyVariableEqualities().walk(query)
    return query


def shatter_constants(query, symbol_table, shatter_all=False):
    head_vars = extract_logic_free_variables(query)
    variable_support = extract_variable_support(query)
    head_vars_choice = set()
    query_for_symbol_sets = query
    for var in head_vars:
        for predicate in variable_support[var]:
            if isinstance(symbol_table[predicate], ProbabilisticChoiceSet):
                head_vars_choice.add(var)
    for var in head_vars_choice:
        query_for_symbol_sets = ExistentialPredicate(
            var, query_for_symbol_sets
        )

    head_vars -= head_vars_choice

    # query = contemplate_for_head_var_equivalence(query, head_vars)

    symbol_sets = sets_per_symbol(query_for_symbol_sets)
    if not shatter_all:
        for k in list(symbol_sets.keys()):
            if not isinstance(symbol_table[k], ProbabilisticFactSet):
                del symbol_sets[k]
    old_keys = set(symbol_table.keys())
    se = ShatterEqualities(symbol_sets, symbol_table, head_vars)
    query = se.walk(RTO.walk(query))
    new_keys = set(symbol_table) - old_keys
    rasc = RelationalAlgebraSelectionConjunction()
    for k in new_keys:
        symbol_table[k] = rasc.walk(symbol_table[k])
    return query



from ...expressions import Symbol
from ...logic import (
    Conjunction,
    Disjunction,
    ExistentialPredicate,
    Implication,
    Negation
)
from .. import transforms


def test_rule_ucq():
    R = Symbol('R')
    S = Symbol('S')
    T = Symbol('T')
    x = Symbol('x')
    y = Symbol('y')
    z = Symbol('z')

    rule = Implication(R(x, y), S(x, y))
    ucq = S(x, y)

    res = transforms.convert_rule_to_ucq(rule)
    assert ucq == res

    rule = Implication(R(x), S(x, y))
    ucq = ExistentialPredicate(y, S(x, y))

    res = transforms.convert_rule_to_ucq(rule)
    assert ucq == res

    rule = Implication(R(x), Conjunction((S(x, y), T(x))))
    ucq = Conjunction((ExistentialPredicate(y, S(x, y)), T(x)))

    res = transforms.convert_rule_to_ucq(rule)
    assert ucq == res

    rule = Implication(R(x), Conjunction((S(x, y), T(z, x))))
    ucq = Conjunction((
        ExistentialPredicate(z, T(z, x)),
        ExistentialPredicate(y, S(x, y)),
    ))

    res = transforms.convert_rule_to_ucq(rule)
    assert ucq == res


def test_unify_existential_variables():
    S = Symbol('S')
    T = Symbol('T')
    x = Symbol('x')
    y = Symbol('y')

    expression = Disjunction((
        ExistentialPredicate(x, T(x)),
        ExistentialPredicate(y, T(y))
    ))

    res = transforms.unify_existential_variables(expression)

    assert res == ExistentialPredicate(y, Disjunction((T(y), T(y))))

    expression = Disjunction((
        Conjunction((S(x, y), ExistentialPredicate(x, T(x)))),
        ExistentialPredicate(y, T(y))
    ))

    res = transforms.unify_existential_variables(expression)

    assert isinstance(res, ExistentialPredicate)
    e_var = res.head
    expected = ExistentialPredicate(
        e_var,
        Disjunction((
            Conjunction((S(x, y), T(e_var))),
            T(e_var)
        ))
    )
    assert res == expected


def test_rule_with_negation_to_ucq():
    R = Symbol('R')
    S = Symbol('S')
    T = Symbol('T')
    x = Symbol('x')
    y = Symbol('y')

    rule = Implication(R(x), Conjunction((S(x, y), Negation(T(y)))))

    expected = ExistentialPredicate(y, Conjunction((S(x, y), Negation(T(y)))))
    res = transforms.convert_rule_to_ucq(rule)

    assert expected == res

    rule = Implication(R(x), Conjunction((S(x, y), S(x, x), Negation(T(y)))))
    expected = Conjunction((
        ExistentialPredicate(y, Conjunction((S(x, y), Negation(T(y))))),
        S(x, x)
    ))
    res = transforms.convert_rule_to_ucq(rule)

    assert expected == res



import logging

from ..datalog import Fact, chase
from ..datalog.negation import DatalogProgramNegation
from ..expressions import Constant, Symbol
from ..logic import Disjunction, Implication, Union
from ..logic.expression_processing import (
    extract_logic_atoms,
    extract_logic_free_variables
)
from ..logic.transformations import (
    MakeExistentialsImplicit,
    PushExistentialsDown,
    RemoveUniversalPredicates,
    convert_to_pnf_with_dnf_matrix
)


LOG = logging.getLogger(__name__)


__all__ = ['is_contained_rule', 'is_contained']


def canonical_database_program(rule):
    '''
    Given a datalog rule R(X):-S_1(X,Y),...,S_N(X,Y) where
    X and Y are sets of symbols, produce the datalog program corresponding
    to the facts
    S_1(A,B).
    S_2(A,B).
    .
    .
    .
    S_N(A,B).
    where A and B are mappings from the symbols in X and Y to constants.
    '''
    consequent, antecedent = rule.unapply()
    return Union(tuple(
        Fact(freeze_atom(atom))
        for atom in extract_logic_atoms(antecedent)
    )), freeze_atom(consequent)


def freeze_atom(atom):
    '''
    Converts all symbols of the atom
    to constants with a string representing
    the symbol.
    '''
    args = (
        Constant(arg.name) if isinstance(arg, Symbol) else arg
        for arg in atom.args
    )
    return atom.functor(*args)


def is_contained_rule(q1, q2):
    '''
    Computes if q1 is contained in q2. Specifically,
    for 2 Datalog rules without constants, computes wether
    q1←q2.
    '''
    s = Symbol.fresh()
    q1 = Implication(
        s(*q1.consequent.args), q1.antecedent
    )
    q2 = Implication(
        s(*q2.consequent.args), q2.antecedent
    )
    d_q2, frozen_head = canonical_database_program(q2)
    dp = DatalogProgramNegation()
    for f in d_q2.formulas:
        dp.walk(f)
    dp.walk(q1)
    solution = chase.Chase(dp).build_chase_solution()
    return frozen_head in solution.as_set()


def is_contained(q1, q2):
    '''
    Computes if q1 is contained in q2. Specifically,
    for 2 non-recursive positive ∃ logic queries,
    without constants, computes wether q1←q2.

    caveat: the free variables of q1 and q2 should be
    the same, else the queries are judge as not contained.
    '''
    s = Symbol.fresh()
    programs = []
    if (
        extract_logic_free_variables(q1) !=
        extract_logic_free_variables(q2)
    ):
        LOG.debug("The free variables are different")
        return False
    for query in (q1, q2):
        program = convert_pos_logic_query_to_datalog_rules(query, s)
        LOG.debug("Canonical program for %s", query)
        programs.append(program)

    for q2_ in programs[1]:
        for q1_ in programs[0]:
            if is_contained_rule(q1_, q2_):
                break
        else:
            return False
    return True


def convert_pos_logic_query_to_datalog_rules(query, head):
    '''
    Converts a positive ∃ logic query without constants
    to a list of datalog rules.
    '''
    ruq = RemoveUniversalPredicates()
    mei = MakeExistentialsImplicit()
    ped = PushExistentialsDown()
    q_args = set(extract_logic_free_variables(query))
    antecedent = ped.walk(ruq.walk(convert_to_pnf_with_dnf_matrix(query)))
    if isinstance(antecedent, Disjunction):
        program = antecedent.formulas
    else:
        program = (antecedent,)
    program = [
        Implication(
            head(*(q_args & extract_logic_free_variables(formula))),
            mei.walk(formula)
        )
        for formula in program
    ]
    return program



"""Module containing the intermediate representation implementation of
relational algebra. It includes, expressions representing the operations,
a solver, to execute the expressions on `RelationalAlgebraSet` objects.
and different static optimisers.
"""

from .optimisers import (
    EliminateTrivialProjections,
    PushInSelections,
    RelationalAlgebraOptimiser,
    RenameOptimizations,
    SimplifyExtendedProjectionsWithConstants
)
from .relational_algebra import (
    BinaryRelationalAlgebraOperation,
    ColumnInt,
    ColumnStr,
    ConcatenateConstantColumn,
    Destroy,
    Difference,
    EquiJoin,
    ExtendedProjection,
    FullOuterNaturalJoin,
    FunctionApplicationListMember,
    GroupByAggregation,
    Intersection,
    LeftNaturalJoin,
    NameColumns,
    NamedRelationalAlgebraFrozenSet,
    NAryRelationalAlgebraOperation,
    NaturalJoin,
    NumberColumns,
    Product,
    Projection,
    RelationalAlgebraColumnInt,
    RelationalAlgebraColumnStr,
    RelationalAlgebraOperation,
    RelationalAlgebraSet,
    RelationalAlgebraSolver,
    RelationalAlgebraStringExpression,
    RenameColumn,
    RenameColumns,
    ReplaceNull,
    Selection,
    UnaryRelationalAlgebraOperation,
    Union,
    eq_,
    get_expression_columns,
    int2columnint_constant,
    str2columnstr_constant
)

__all__ = [
    "BinaryRelationalAlgebraOperation",
    "ColumnInt",
    "ColumnStr",
    "ConcatenateConstantColumn",
    "Destroy",
    "Difference",
    "EliminateTrivialProjections",
    "eq_",
    "EquiJoin",
    "ExtendedProjection",
    "FullOuterNaturalJoin",
    "FunctionApplicationListMember",
    "get_expression_columns",
    "GroupByAggregation",
    "int2columnint_constant",
    "Intersection",
    "LeftNaturalJoin",
    "NameColumns",
    "NamedRelationalAlgebraFrozenSet",
    "NAryRelationalAlgebraOperation",
    "NaturalJoin",
    "NumberColumns",
    "Product",
    "Projection",
    "RelationalAlgebraColumnInt",
    "RelationalAlgebraColumnStr",
    "RelationalAlgebraOperation",
    "RelationalAlgebraOptimiser",
    "PushInSelections",
    "RelationalAlgebraSet",
    "RelationalAlgebraSolver",
    "RelationalAlgebraStringExpression",
    "RenameColumn",
    "RenameColumns",
    "RenameOptimizations",
    "ReplaceNull",
    "Selection",
    "SimplifyExtendedProjectionsWithConstants",
    "str2columnstr_constant",
    "UnaryRelationalAlgebraOperation",
    "Union",
]



import time
from concurrent.futures import Future, TimeoutError
from unittest.mock import create_autospec
from uuid import uuid4

import pytest
from neurolang.frontend.probabilistic_frontend import NeurolangPDL
from ..app import NeurolangQueryManager
from ..engines import NeurolangEngineConfiguration
from ....exceptions import UnexpectedTokenError


@pytest.fixture
def conf():
    conf = create_autospec(NeurolangEngineConfiguration)
    conf.key = "mockengine"
    conf.create.side_effect = lambda : NeurolangPDL()
    return conf

def test_nqm_creates_engines(conf):
    opts = {conf: 5}
    nqm = NeurolangQueryManager(opts)
    time.sleep(.5)

    assert conf.key in nqm.engines
    assert nqm.engines[conf.key].counter == 5
    assert len(nqm.engines[conf.key].engines) == 5
    assert conf.create.call_count == 5

def test_nqm_submits_queries(conf):
    opts = {conf: 2}
    nqm = NeurolangQueryManager(opts)
    time.sleep(.2)

    uuid = uuid4()
    query = "invalid query"
    res = nqm.submit_query(uuid, query, conf.key)
    assert isinstance(res, Future)
    assert res is nqm.get_result(uuid)

    # wait for future execution. It should raise the exception
    with pytest.raises(UnexpectedTokenError):
        res.result()

def test_nqm_waits_for_engines_to_be_available(conf):
    # create two engines
    opts = {conf: 2}
    nqm = NeurolangQueryManager(opts)
    time.sleep(.2)

    # acquire the two engines manually to block execution
    nqm.engines[conf.key].sema.acquire()
    nqm.engines[conf.key].sema.acquire()

    # submit three tasks
    id1 = uuid4()
    id2 = uuid4()
    id3 = uuid4()
    res1 = nqm.submit_query(id1, "A('x', 3)", conf.key)
    res2 = nqm.submit_query(id2, "A('y', 2)", conf.key)
    res3 = nqm.submit_query(id3, "A('z', 1)", conf.key)
    assert isinstance(res1, Future)
    assert isinstance(res2, Future)
    assert isinstance(res3, Future)
    
    # wait 1sec for 1st query to complete
    with pytest.raises(TimeoutError):
        res1.result(timeout=1)
    # q1 & q2 should be running but not done. q3 should not be running.
    assert not res1.done()
    assert not res2.done()
    assert res1.running()
    assert res2.running()
    assert not res3.running()

    # release an engine
    nqm.engines[conf.key].sema.release()
    # all 3 queries should complete
    res1 = nqm.get_result(id1)
    res2 = nqm.get_result(id2)
    res3 = nqm.get_result(id3)
    assert res1.result() is not None
    assert res2.result() is not None
    assert res3.result() is not None

def test_nqm_cancels_queries(conf):
    # create one engine
    opts = {conf: 1}
    nqm = NeurolangQueryManager(opts)
    time.sleep(.1)

    # acquire the engine manually to block execution
    nqm.engines[conf.key].sema.acquire()

    # submit two tasks
    id1 = uuid4()
    id2 = uuid4()
    res1 = nqm.submit_query(id1, "A('x', 3)", conf.key)
    res2 = nqm.submit_query(id2, "A('y', 2)", conf.key)
    time.sleep(.1)
    assert not res1.done()
    assert res1.running()
    assert not res2.running()

    # cancel the second task
    cancelled = nqm.cancel(id2)
    time.sleep(.1)
    assert cancelled
    res2 = nqm.get_result(id2)
    assert not res2.running()
    assert res2.cancelled()

    # release engine to finish first task
    nqm.engines[conf.key].sema.release()
    assert res1.result() is not None



import typing

import pytest

from ...exceptions import ForbiddenDisjunctionError
from ...expressions import Constant, FunctionApplication, Symbol
from ...logic import Implication, Union
from ..cplogic.program import CPLogicProgram
from ..expressions import ProbabilisticFact
from ..query_resolution import QueryBasedProbFactToDetRule

P = Symbol("P")
Q = Symbol("Q")
x = Symbol("x")
y = Symbol("y")
p = Symbol("p")


class QueryBasedProbFactToDetRuleProgramTest(
    QueryBasedProbFactToDetRule,
    CPLogicProgram,
):
    pass


def test_query_based_pfact():
    """
    Translate rule
        P(x) : (p / 2) :- Q(x)
    to the two rules
        _f1_(_f2_, x) :- Q(x), _f2_ = (p / 2)
        P(x) : _f2_ :- _f1_(_f2_, x)

    """
    pfact = Implication(
        ProbabilisticFact(p / Constant(2), P(x)),
        Q(x, p),
    )
    program = QueryBasedProbFactToDetRuleProgramTest()
    program.walk(Union((pfact,)))
    program_rules = list()
    for union in program.intensional_database().values():
        program_rules += union.formulas
    det_rule = next(
        formula
        for formula in program_rules
        if (
            isinstance(formula.consequent, FunctionApplication)
            # check _f1_
            and formula.consequent.functor.is_fresh
            # check _f2_ first arg in consequent
            and formula.consequent.args[0].is_fresh
            and formula.consequent.args[1:] == (x,)
        )
    )
    assert any(
        isinstance(formula.consequent, ProbabilisticFact)
        and formula.consequent.probability == det_rule.consequent.args[0]
        and formula.antecedent == det_rule.consequent
        and formula.consequent.body == P(x)
        for formula in program_rules
    )


def test_prevent_combination_of_query_based_and_set_based():
    pfact = Implication(
        ProbabilisticFact(p / Constant(2), P(x)),
        Q(x, p),
    )
    cpl = QueryBasedProbFactToDetRuleProgramTest()
    cpl.add_probabilistic_facts_from_tuples(P, [(0.2, "a")])
    with pytest.raises(ForbiddenDisjunctionError):
        cpl.walk(pfact)


def test_callable_symbol_probability():
    fun = Symbol[typing.Callable[[float], float]]("fun")
    cpl = QueryBasedProbFactToDetRuleProgramTest()
    cpl.symbol_table[fun] = Constant(lambda x: x ** 2)
    pfact = Implication(
        ProbabilisticFact(fun(x), P(y)),
        Q(x, y),
    )
    cpl.walk(Union((pfact,)))
    rules = list()
    for union in cpl.intensional_database().values():
        rules += union.formulas
    det_rule = next(
        formula
        for formula in rules
        if (
            isinstance(formula.consequent, FunctionApplication)
            # check _f1_
            and formula.consequent.functor.is_fresh
            # check _f2_ first arg in consequent
            and formula.consequent.args[0].is_fresh
            and formula.consequent.args[1:] == (y,)
        )
    )
    assert any(
        isinstance(formula.consequent, ProbabilisticFact)
        and formula.consequent.probability == det_rule.consequent.args[0]
        and formula.antecedent == det_rule.consequent
        and formula.consequent.body == P(y)
        for formula in rules
    )



from ...expressions import Symbol
from .. import (
    Conjunction,
    Disjunction,
    ExistentialPredicate,
    Negation,
    UniversalPredicate
)
from ..transformations import (
    PushExistentialsDown,
    PushUniversalsDown
)

P = Symbol("P")
Q = Symbol("Q")
R = Symbol("R")
X = Symbol("X")
Y = Symbol("Y")
Z = Symbol("Z")


PED = PushExistentialsDown()
PUD = PushUniversalsDown()


def test_push_existentials_down():
    exp = ExistentialPredicate(
        X,
        Disjunction((P(X), Q(X)))
    )

    assert PED.walk(exp) == Disjunction((
        ExistentialPredicate(X, P(X)),
        ExistentialPredicate(X, Q(X))
    ))

    exp = ExistentialPredicate(
        X, Conjunction((
            P(X), Q(X)
        ))
    )

    assert PED.walk(exp) == exp

    exp = ExistentialPredicate(
        X, Conjunction((
            P(Y), Q(X)
        ))
    )

    assert PED.walk(exp) == Conjunction((
        ExistentialPredicate(X, Q(X)),
        P(Y)
    ))

    exp = ExistentialPredicate(
        X, Conjunction((
            P(Y), Q(X)
        ))
    )

    assert PED.walk(exp) == Conjunction((
        ExistentialPredicate(X, Q(X)),
        P(Y)
    ))

    exp = ExistentialPredicate(
        X, Conjunction((
            P(Y, Z), Negation(Q(X, Z))
        ))
    )

    assert PED.walk(exp) == exp

    exp = ExistentialPredicate(
        X,
        ExistentialPredicate(
            Y,
            Conjunction((
                P(X), Q(Y)
            ))
        )
    )

    assert PED.walk(exp) == Conjunction((
        ExistentialPredicate(X, P(X)),
        ExistentialPredicate(Y, Q(Y))
    ))

    exp = ExistentialPredicate(
        X,
        UniversalPredicate(
            Y,
            Conjunction((P(X), Q(Y)))
        )
    )

    assert PED.walk(exp) == exp


def test_push_universals_down():
    exp = UniversalPredicate(
        X,
        Conjunction((P(X), Q(X)))
    )

    assert PUD.walk(exp) == Conjunction((
        UniversalPredicate(X, P(X)),
        UniversalPredicate(X, Q(X))
    ))

    exp = UniversalPredicate(
        X, Disjunction((
            P(X), Q(X)
        ))
    )

    assert PUD.walk(exp) == exp

    exp = UniversalPredicate(
        X, Disjunction((
            P(Y), Q(X)
        ))
    )

    assert PUD.walk(exp) == Disjunction((
        UniversalPredicate(X, Q(X)),
        P(Y)
    ))

    exp = UniversalPredicate(
        X, Disjunction((
            P(Y), Q(X)
        ))
    )

    assert PUD.walk(exp) == Disjunction((
        UniversalPredicate(X, Q(X)),
        P(Y)
    ))

    exp = UniversalPredicate(
        X, Conjunction((
            P(Y, Z), Negation(Q(X, Z))
        ))
    )

    assert PUD.walk(exp) == Conjunction((
        P(Y, Z),
        UniversalPredicate(X, Negation(Q(X, Z)))
    ))

    exp = UniversalPredicate(
        X,
        UniversalPredicate(
            Y,
            Conjunction((
                P(X), Q(Y)
            ))
        )
    )

    assert PUD.walk(exp) == Conjunction((
        UniversalPredicate(X, P(X)),
        UniversalPredicate(Y, Q(Y))
    ))

    exp = UniversalPredicate(
        X,
        ExistentialPredicate(
            Y,
            Conjunction((P(X), Q(Y)))
        )
    )

    assert PUD.walk(exp) == exp
